# Copyright (c) 2016 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import unittest

try:  # noqa: C901
    import gurobipy
except ImportError as e:

    class TestMissingDependency(unittest.TestCase):

        @unittest.skip('Missing dependency - ' + str(e))
        def test_fail(self):
            pass

else:

    import copy

    import random

    import os

    import nose
    import pickle

    from optlang.gurobi_interface import Variable, Constraint, Model, Objective
    from gurobipy import GurobiError
    from optlang.tests import abstract_test_cases
    from optlang import gurobi_interface

    random.seed(666)
    TESTMODELPATH = os.path.join(os.path.dirname(__file__), 'data/model.lp')
    TESTMILPMODELPATH = os.path.join(os.path.dirname(__file__), 'data/simple_milp.lp')
    CONVEX_QP_PATH = os.path.join(os.path.dirname(__file__), 'data/qplib_3256.lp')
    NONCONVEX_QP_PATH = os.path.join(os.path.dirname(__file__), 'data/qplib_1832.lp')


    class VariableTestCase(abstract_test_cases.AbstractVariableTestCase):
        interface = gurobi_interface

        def test_internal_variable(self):
            self.assertEqual(self.var._internal_variable, None)

        def test_gurobi_change_name(self):
            self.model.add(self.var)
            self.model.update()
            self.var.name = "test_2"
            self.assertEqual(self.var._internal_variable.getAttr("VarName"), "test_2")

        def test_get_primal(self):
            self.assertEqual(self.var.primal, None)
            model = Model(problem=gurobipy.read(TESTMODELPATH))
            model.optimize()
            for i, j in zip([var.primal for var in model.variables],
                            [0.8739215069684306, -16.023526143167608, 16.023526143167604, -14.71613956874283,
                             14.71613956874283, 4.959984944574658, 4.959984944574657, 4.959984944574658,
                             3.1162689467973905e-29, 2.926716099010601e-29, 0.0, 0.0, -6.112235045340358e-30,
                             -5.6659435396316186e-30, 0.0, -4.922925402711085e-29, 0.0, 9.282532599166613, 0.0,
                             6.00724957535033, 6.007249575350331, 6.00724957535033, -5.064375661482091,
                             1.7581774441067828, 0.0, 7.477381962160285, 0.0, 0.22346172933182767, 45.514009774517454,
                             8.39, 0.0, 6.007249575350331, 0.0, -4.541857463865631, 0.0, 5.064375661482091, 0.0, 0.0,
                             2.504309470368734, 0.0, 0.0, -22.809833310204958, 22.809833310204958, 7.477381962160285,
                             7.477381962160285, 1.1814980932459636, 1.496983757261567, -0.0, 0.0, 4.860861146496815,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 5.064375661482091, 0.0, 5.064375661482091, 0.0, 0.0,
                             1.496983757261567, 10.000000000000002, -10.0, 0.0, 0.0, 0.0, 0.0, 0.0, -29.175827135565804,
                             43.598985311997524, 29.175827135565804, 0.0, 0.0, 0.0, -1.2332237321082153e-29,
                             3.2148950476847613, 38.53460965051542, 5.064375661482091, 0.0, -1.2812714099825612e-29,
                             -1.1331887079263237e-29, 17.530865429786694, 0.0, 0.0, 0.0, 4.765319193197458,
                             -4.765319193197457, 21.79949265599876, -21.79949265599876, -3.2148950476847613, 0.0,
                             -2.281503094067127, 2.6784818505075303, 0.0]):
                self.assertAlmostEqual(i, j)

        def test_changing_variable_names_is_reflected_in_the_solver(self):
            model = Model(problem=gurobipy.read(TESTMODELPATH))
            for i, variable in enumerate(model.variables):
                print(variable._internal_variable is not None)
                print(variable.problem.name)
                variable.name = "var" + str(i)
                print(variable.problem.name)
                print(variable.name)
                print(variable._internal_variable is not None)
                self.assertEqual(variable.name, "var" + str(i))
                self.assertEqual(variable._internal_variable.getAttr('VarName'), "var" + str(i))

        def test_gurobi_setting_bounds(self):
            var = self.var
            model = self.model
            model.add(var)
            model.update()
            var.lb = 1
            self.assertEqual(var.lb, 1)
            model.problem.update()
            self.assertEqual(var._internal_variable.getAttr('LB'), 1)
            var.ub = 2
            self.assertEqual(var.ub, 2)
            model.problem.update()
            self.assertEqual(var._internal_variable.getAttr('UB'), 2)


    class ConstraintTestCase(abstract_test_cases.AbstractConstraintTestCase):
        interface = gurobi_interface

        def test_get_primal(self):
            self.assertEqual(self.constraint.primal, None)
            self.model.optimize()
            print([constraint.primal for constraint in self.model.constraints])
            for i, j in zip([constraint.primal for constraint in self.model.constraints],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             4.048900234729145e-15, 0.0, 0.0, 0.0, -3.55971196577979e-16, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 2.5546369406238147e-17, 0.0, -5.080374405378186e-29, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]):
                self.assertAlmostEqual(i, j)


    class ObjectiveTestCase(abstract_test_cases.AbstractObjectiveTestCase):
        interface = gurobi_interface

        def setUp(self):
            problem = gurobipy.read(TESTMODELPATH)
            self.model = Model(problem=problem)
            self.obj = self.model.objective

        def test_change_direction(self):
            self.obj.direction = "min"
            self.assertEqual(self.obj.direction, "min")
            self.assertEqual(self.model.problem.getAttr('ModelSense'), gurobipy.GRB.MAXIMIZE)
            self.model.update()
            self.assertEqual(self.model.problem.getAttr('ModelSense'), gurobipy.GRB.MINIMIZE)

            self.obj.direction = "max"
            self.assertEqual(self.obj.direction, "max")
            self.assertEqual(self.model.problem.getAttr('ModelSense'), gurobipy.GRB.MINIMIZE)
            self.model.update()
            self.assertEqual(self.model.problem.getAttr('ModelSense'), gurobipy.GRB.MAXIMIZE)


    class ConfigurationTestCase(abstract_test_cases.AbstractConfigurationTestCase):
        interface = gurobi_interface


    class ModelTestCase(abstract_test_cases.AbstractModelTestCase):
        interface = gurobi_interface

        def test_gurobi_create_empty_model(self):
            model = Model()
            self.assertEqual(model.problem.getAttr('NumVars'), 0)
            self.assertEqual(model.problem.getAttr('NumConstrs'), 0)
            self.assertEqual(model.name, None)
            self.assertEqual(model.problem.getAttr('ModelName'), '')
            model = Model(name="empty_problem")
            self.assertEqual(model.problem.getAttr('ModelName'), 'empty_problem')

        def test_pickle_ability(self):
            self.model.optimize()
            value = self.model.objective.value
            pickle_string = pickle.dumps(self.model)
            from_pickle = pickle.loads(pickle_string)
            from_pickle.optimize()
            self.assertAlmostEqual(value, from_pickle.objective.value)
            self.assertEqual([(var.lb, var.ub, var.name, var.type) for var in from_pickle.variables.values()],
                             [(var.lb, var.ub, var.name, var.type) for var in self.model.variables.values()])
            self.assertEqual([(constr.lb, constr.ub, constr.name) for constr in from_pickle.constraints],
                             [(constr.lb, constr.ub, constr.name) for constr in self.model.constraints])

        def test_config_gets_copied_too(self):
            self.assertEqual(self.model.configuration.verbosity, 0)
            self.model.configuration.verbosity = 3
            model_copy = copy.copy(self.model)
            self.assertEqual(model_copy.configuration.verbosity, 3)

        def test_init_from_existing_problem(self):
            self.assertEqual(len(self.model.variables), len(self.model.problem.getVars()))
            self.assertEqual(len(self.model.constraints), len(self.model.problem.getConstrs()))
            self.assertEqual(self.model.variables.keys(),
                             [var.VarName for var in self.model.problem.getVars()])

            self.assertEqual(self.model.constraints.keys(),
                             [constr.ConstrName for constr in self.model.problem.getConstrs()])

        def test_gurobi_add_variable(self):
            var = Variable('x')
            self.model.add(var)
            print(self.model._pending_modifications)
            self.assertTrue(var in self.model.variables.values())
            self.assertEqual(self.model.variables.values().count(var), 1)
            self.assertEqual(self.model.variables['x'].problem, var.problem)
            print(var.name)
            print(self.model.problem.getVars())
            print(self.model._pending_modifications)
            self.model.update()
            print(self.model._pending_modifications)
            print(self.model.problem.getVars())
            self.assertEqual(self.model.problem.getVarByName(var.name).getAttr('VType'), gurobipy.GRB.CONTINUOUS)
            var = Variable('y', lb=-13)
            self.model.add(var)
            self.assertTrue(var in self.model.variables.values())
            self.model.problem.update()
            self.assertEqual(self.model.problem.getVarByName(var.name).getAttr('VType'), gurobipy.GRB.CONTINUOUS)
            self.assertEqual(self.model.variables['x'].lb, None)
            self.assertEqual(self.model.variables['x'].ub, None)
            self.assertEqual(self.model.variables['y'].lb, -13)
            self.assertEqual(self.model.variables['x'].ub, None)
            var = Variable('x_with_ridiculously_long_variable_name_asdffffffffasdfasdfasdfasdfasdfasdfasdf')
            self.model.add(var)
            self.assertTrue(var in self.model.variables.values())
            self.assertEqual(self.model.variables.values().count(var), 1)

        def test_gurobi_add_integer_var(self):
            var = Variable('int_var', lb=-13, ub=500, type='integer')
            self.model.add(var)
            self.assertEqual(self.model.variables['int_var'].type, 'integer')
            self.assertEqual(self.model.problem.getVarByName(var.name).getAttr('VType'), gurobipy.GRB.INTEGER)
            self.assertEqual(self.model.variables['int_var'].ub, 500)
            self.assertEqual(self.model.variables['int_var'].lb, -13)

        def test_add_non_cplex_conform_variable(self):
            var = Variable('12x!!@#5_3', lb=-666, ub=666)
            self.model.add(var)
            self.assertTrue(var in self.model.variables.values())
            self.model.problem.update()
            self.assertEqual(var.name, self.model.problem.getVarByName(var.name).VarName)
            self.assertEqual(self.model.variables['12x!!@#5_3'].lb, -666)
            self.assertEqual(self.model.variables['12x!!@#5_3'].ub, 666)
            repickled = pickle.loads(pickle.dumps(self.model))
            var_from_pickle = repickled.variables['12x!!@#5_3']
            self.assertEqual(var_from_pickle.name, repickled.problem.getVarByName(var.name).VarName)

        def test_gurobi_add_constraints(self):
            x = Variable('x', lb=0, ub=1, type='binary')
            y = Variable('y', lb=-181133.3, ub=12000., type='continuous')
            z = Variable('z', lb=0., ub=10., type='integer')
            constr1 = Constraint(0.3 * x + 0.4 * y + 66. * z, lb=-100, ub=0., name='test')
            constr2 = Constraint(2.333 * x + y + 3.333, ub=100.33, name='test2')
            constr3 = Constraint(2.333 * x + y + z, lb=-300)
            constr4 = Constraint(x, lb=-300, ub=-300)
            self.model.add(constr1)
            self.model.add(constr2)
            self.model.add(constr3)
            self.model.add(constr4)
            self.model.problem.update()
            self.assertIn(constr1.name, self.model.constraints)
            self.assertIn(constr2.name, self.model.constraints)
            self.assertIn(constr3.name, self.model.constraints)
            self.assertIn(constr4.name, self.model.constraints)
            # constr1
            coeff_dict = dict()
            internal_constraint = self.model.problem.getConstrByName(constr1.name)
            row = self.model.problem.getRow(internal_constraint)
            for i in range(row.size()):
                coeff_dict[row.getVar(i).VarName] = row.getCoeff(i)
            self.assertDictEqual(coeff_dict, {'x': 0.3, 'y': 0.4, 'z': 66., 'test_aux': -1.0})
            self.assertEqual(internal_constraint.RHS, constr1.lb)
            self.assertEqual(self.model.problem.getVarByName(internal_constraint.getAttr('ConstrName') + '_aux'), 100)
            # constr2
            coeff_dict = dict()
            internal_constraint = self.model.problem.getConstrByName(constr2.name)
            row = self.model.problem.getRow(internal_constraint)
            for i in range(row.size()):
                coeff_dict[row.getVar(i).VarName] = row.getCoeff(i)
            self.assertDictEqual(coeff_dict, {'x': 2.333, 'y': 1.})
            self.assertEqual(internal_constraint.RHS, constr2.ub)
            self.assertEqual(internal_constraint.Sense, '<')
            # constr3
            coeff_dict = dict()
            internal_constraint = self.model.problem.getConstrByName(constr3.name)
            print(internal_constraint)
            row = self.model.problem.getRow(internal_constraint)
            for i in range(row.size()):
                coeff_dict[row.getVar(i).VarName] = row.getCoeff(i)
            self.assertDictEqual(coeff_dict, {'x': 2.333, 'y': 1., 'z': 1.})
            self.assertEqual(internal_constraint.RHS, constr3.lb)
            self.assertEqual(internal_constraint.Sense, '>')
            # constr4
            coeff_dict = dict()
            internal_constraint = self.model.problem.getConstrByName(constr4.name)
            print(internal_constraint)
            row = self.model.problem.getRow(internal_constraint)
            for i in range(row.size()):
                coeff_dict[row.getVar(i).VarName] = row.getCoeff(i)
            self.assertDictEqual(coeff_dict, {'x': 1})
            self.assertEqual(internal_constraint.RHS, constr4.lb)
            self.assertEqual(internal_constraint.Sense, '=')

        def test_change_of_constraint_is_reflected_in_low_level_solver(self):
            x = Variable('x', lb=-83.3, ub=1324422.)
            y = Variable('y', lb=-181133.3, ub=12000.)
            constraint = Constraint(0.3 * x + 0.4 * y, lb=-100, name='test')
            self.assertEqual(constraint._internal_constraint, None)
            self.model.add(constraint)
            self.assertEqual(self.model.constraints['test'].lb, -100)
            self.assertEqual(
                (self.model.constraints['test'].expression - (0.4 * y + 0.3 * x)).expand() - 0,
                0
            )
            z = Variable('z', lb=3, ub=10, type='integer')
            self.assertEqual(z._internal_variable, None)
            constraint += 77. * z
            self.assertEqual(z._internal_variable, self.model.problem.getVarByName('z'))
            self.assertEqual(self.model.constraints['test'].lb, -100)
            self.assertEqual(
                (self.model.constraints['test'].expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand() - 0,
                0
            )

        def test_constraint_set_problem_to_None_caches_the_latest_expression_from_solver_instance(self):
            x = Variable('x', lb=-83.3, ub=1324422.)
            y = Variable('y', lb=-181133.3, ub=12000.)
            constraint = Constraint(0.3 * x + 0.4 * y, lb=-100, name='test')
            self.model.add(constraint)
            z = Variable('z', lb=2, ub=5, type='integer')
            constraint += 77. * z
            self.model.remove(constraint)
            self.assertEqual(constraint.lb, -100)
            self.assertEqual(
                (constraint.expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand() - 0, 0
            )

        def test_change_of_objective_is_reflected_in_low_level_solver(self):
            x = Variable('x', lb=-83.3, ub=1324422.)
            y = Variable('y', lb=-181133.3, ub=12000.)
            objective = Objective(0.3 * x + 0.4 * y, name='test', direction='max')
            self.model.objective = objective
            self.model.update()
            grb_obj = self.model.problem.getObjective()
            grb_x = self.model.problem.getVarByName(x.name)
            grb_y = self.model.problem.getVarByName(y.name)
            expected = {grb_x: 0.3, grb_y: 0.4}
            for i in range(grb_obj.size()):
                self.assertEqual(grb_obj.getCoeff(i), expected[grb_obj.getVar(i)])
            z = Variable('z', lb=4, ub=4, type='integer')
            grb_z = self.model.problem.getVarByName(z.name)
            self.model.objective += 77. * z
            expected[grb_z] = 77.
            self.model.update()
            for i in range(grb_obj.size()):
                self.assertEqual(grb_obj.getCoeff(i), expected[grb_obj.getVar(i)])

        def test_change_variable_bounds(self):
            inner_prob = self.model.problem
            inner_problem_bounds = [(variable.LB, variable.UB) for variable in inner_prob.getVars()]
            bounds = [(var.lb, var.ub) for var in self.model.variables.values()]
            self.assertEqual(bounds, inner_problem_bounds)
            for var in self.model.variables.values():
                var.lb = random.uniform(-1000, 1000)
                var.ub = random.uniform(var.lb, 1000)
            self.model.update()
            inner_problem_bounds_new = [(variable.LB, variable.UB) for variable in inner_prob.getVars()]
            bounds_new = [(var.lb, var.ub) for var in self.model.variables.values()]
            self.assertNotEqual(bounds, bounds_new)
            self.assertNotEqual(inner_problem_bounds, inner_problem_bounds_new)
            self.assertEqual(bounds_new, inner_problem_bounds_new)

        def test_gurobi_change_variable_type(self):
            for variable in self.model.variables:
                variable.type = 'integer'
            self.model.update()
            for variable in self.model.problem.getVars():
                self.assertEqual(variable.VType, gurobipy.GRB.INTEGER)

        def test_change_constraint_bounds(self):
            inner_prob = self.model.problem
            inner_problem_bounds = []
            for constr in inner_prob.getConstrs():
                aux_var = inner_prob.getVarByName(constr.getAttr('ConstrName') + '_aux')
                if aux_var is None:
                    inner_problem_bounds.append((constr.RHS, constr.RHS))
                else:
                    inner_problem_bounds.append((aux_var.UB, constr.RHS))
            print(len(self.model.constraints))
            print(len(self.model.problem.getConstrs()))
            bounds = [(constr.lb, constr.ub) for constr in self.model.constraints]
            print('bounds', inner_problem_bounds)
            print('bounds', bounds)
            self.assertEqual(bounds, inner_problem_bounds)

        @unittest.skip('Not supported yet')
        def test_iadd_objective(self):
            v2, v3 = self.model.variables.values()[1:3]
            print(v2, v3)
            # 1/0
            self.model.objective += 2. * v2 - 3. * v3
            internal_objective = self.model.problem.getObjective()
            result = {}
            for i in range(internal_objective.size()):
                var = internal_objective.getVar(i)
                coeff = internal_objective.getCoeff(i)
                result[var.VarName] = coeff
            self.assertDictEqual(result, {'R_Biomass_Ecoli_core_w_GAM': 1.0})
            self.model.update()
            self.assertDictEqual(result, {'R_Biomass_Ecoli_core_w_GAM': 1.0, 'R_PGK': 2, 'R_GAPD': -3})

        @unittest.skip('Not supported yet')
        def test_imul_objective(self):
            self.model.objective *= 2.
            obj_coeff = list()
            self.assertEqual(obj_coeff,
                             [0.0, 2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                              0.0,
                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                              0.0,
                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                              0.0,
                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                              0.0,
                              0.0]
                             )

        def test_set_copied_objective(self):
            obj_copy = copy.copy(self.model.objective)
            self.model.objective = obj_copy
            self.assertEqual(self.model.objective.direction, "max")
            self.assertEqual(
                (self.model.objective.expression - (1.0 * self.model.variables.R_Biomass_Ecoli_core_w_GAM)).expand() - 0,
                0
            )

        def test_timeout(self):
            self.model.configuration.timeout = 0
            status = self.model.optimize()
            print(status)
            self.assertEqual(status, 'time_limit')

        def test_set_linear_coefficients_objective(self):
            self.model.objective.set_linear_coefficients({self.model.variables.R_TPI: 666.})
            self.model.update()
            grb_obj = self.model.problem.getObjective()
            for i in range(grb_obj.size()):
                if 'R_TPI' == grb_obj.getVar(i).getAttr('VarName'):
                    self.assertEqual(grb_obj.getCoeff(i), 666.)

        def test_set_linear_coefficients_constraint(self):
            constraint = self.model.constraints.M_atp_c
            constraint.set_linear_coefficients({self.model.variables.R_Biomass_Ecoli_core_w_GAM: 666.})
            self.model.update()
            row = self.model.problem.getRow(self.model.problem.getConstrByName(constraint.name))
            for i in range(row.size()):
                col_name = row.getVar(i).getAttr('VarName')
                if col_name == 'R_Biomass_Ecoli_core_w_GAM':
                    self.assertEqual(row.getCoeff(i), 666.)


    class QuadraticProgrammingTestCase(abstract_test_cases.AbstractQuadraticProgrammingTestCase):
            def setUp(self):
                self.model = Model()
                self.x1 = Variable("x1", lb=0)
                self.x2 = Variable("x2", lb=0)
                self.c1 = Constraint(self.x1 + self.x2, lb=1)
                self.model.add([self.x1, self.x2, self.c1])

            def test_convex_obj(self):
                model = self.model
                obj = Objective(self.x1 ** 2 + self.x2 ** 2, direction="min")
                model.objective = obj
                model.optimize()
                self.assertAlmostEqual(model.objective.value, 0.5)
                self.assertAlmostEqual(self.x1.primal, 0.5)
                self.assertAlmostEqual(self.x2.primal, 0.5)

                obj_2 = Objective(self.x1, direction="min")
                model.objective = obj_2
                model.optimize()
                self.assertAlmostEqual(model.objective.value, 0.0)
                self.assertAlmostEqual(self.x1.primal, 0.0)
                self.assertGreaterEqual(self.x2.primal, 1.0)

            # According to documentation and mailing lists Gurobi cannot solve non-convex QP
            # However version 7.0 solves this fine. Skipping for now
            @unittest.skip("Can gurobi solve non-convex QP?")
            def test_non_convex_obj(self):
                model = self.model
                obj = Objective(self.x1 * self.x2, direction="min")
                model.objective = obj
                self.assertRaises(GurobiError, model.optimize)

                obj_2 = Objective(self.x1, direction="min")
                model.objective = obj_2
                model.optimize()
                self.assertAlmostEqual(model.objective.value, 0.0)
                self.assertAlmostEqual(self.x1.primal, 0.0)
                self.assertGreaterEqual(self.x2.primal, 1.0)

            def test_qp_convex(self):
                model = Model(problem=gurobipy.read(CONVEX_QP_PATH))
                self.assertEqual(len(model.variables), 651)
                self.assertEqual(len(model.constraints), 501)
                for constraint in model.constraints:
                    self.assertTrue(constraint.is_Linear, "%s should be linear" % (str(constraint.expression)))
                    self.assertFalse(constraint.is_Quadratic, "%s should not be quadratic" % (str(constraint.expression)))

                self.assertTrue(model.objective.is_Quadratic, "objective should be quadratic")
                self.assertFalse(model.objective.is_Linear, "objective should not be linear")

                model.optimize()
                self.assertAlmostEqual(model.objective.value, 32.2291282)

            @unittest.skip("Takes a very long time")
            def test_qp_non_convex(self):
                model = Model(problem=gurobipy.read(NONCONVEX_QP_PATH))
                self.assertEqual(len(model.variables), 31)
                self.assertEqual(len(model.constraints), 1)
                for constraint in model.constraints:
                    self.assertTrue(constraint.is_Linear, "%s should be linear" % (str(constraint.expression)))
                    self.assertFalse(constraint.is_Quadratic, "%s should not be quadratic" % (str(constraint.expression)))

                self.assertTrue(model.objective.is_Quadratic, "objective should be quadratic")
                self.assertFalse(model.objective.is_Linear, "objective should not be linear")

                self.assertRaises(GurobiError, model.optimize)

            def test_quadratic_objective_expression(self):
                objective = Objective(self.x1 ** 2 + self.x2 ** 2, direction="min")
                self.model.objective = objective
                self.assertEqual((self.model.objective.expression - (self.x1 ** 2 + self.x2 ** 2)).simplify(), 0)

if __name__ == '__main__':
    nose.runmodule()



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from optlang.glpk_interface import Model, Variable, Constraint, Objective
from optlang.duality import convert_linear_problem_to_dual
import optlang


class DualityTestCase(unittest.TestCase):
    def setUp(self):
        x = Variable("x", lb=0)
        y = Variable("y", lb=0)
        z = Variable("z", lb=0)

        c1 = Constraint(x, ub=10)
        c2 = Constraint(y - 2 * z, lb=0, ub=0)
        c3 = Constraint(x - z, lb=3, ub=15)

        obj = Objective(x + y + z)

        model = Model()
        model.add([c1, c2, c3])
        model.objective = obj
        self.model = model
        self.c1 = c1

    def test_dual_value_equal_primal(self):
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 31)
        self.assertEqual(dual.objective.value, 31)

    def test_making_primal_unbounded(self):
        self.model.remove(self.c1)
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()
        self.assertEqual(primal_status, optlang.interface.UNBOUNDED)
        self.assertEqual(dual_status, optlang.interface.INFEASIBLE)

    def test_making_primal_infeasible(self):
        # Note: The dual of an infeasible primal will be either unbounded or infeasible. In this case it is unbounded.
        c4 = Constraint(self.model.variables["z"], lb=100)
        self.model.add(c4)
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()
        self.assertEqual(primal_status, optlang.interface.INFEASIBLE)
        self.assertEqual(dual_status, optlang.interface.UNBOUNDED)

    def test_non_standard_raises(self):
        self.model.variables[0].lb = -10
        self.assertRaises(ValueError, convert_linear_problem_to_dual, self.model)

    def test_non_linear_raises(self):
        model = optlang.interface.Model.clone(self.model)  # GLPK interface doesn't allow non-linear constraints
        c4 = optlang.interface.Constraint(2 * model.variables["x"]**2, ub=7)
        model.add(c4)
        self.assertRaises(ValueError, convert_linear_problem_to_dual, model)

    def test_non_continuous_raises(self):
        self.model.variables["z"].type = "integer"
        self.assertRaises(ValueError, convert_linear_problem_to_dual, self.model)

    def test_infinity(self):
        dual = convert_linear_problem_to_dual(self.model, infinity=1000)
        for var in dual.variables:
            self.assertTrue(var.ub <= 1000)
            self.assertTrue(var.lb >= -1000)

    def test_dual_is_standard_form(self):
        dual = convert_linear_problem_to_dual(self.model, infinity=1000)
        dual_2 = convert_linear_problem_to_dual(self.model, maintain_standard_form=False, infinity=1000)
        self.assertTrue(all(var.lb >= 0 for var in dual.variables))
        self.assertFalse(all(var.lb >= 0 for var in dual_2.variables))

    def test_variable_with_positive_lower_bound(self):
        self.model.variables["x"].lb = 1
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 31)
        self.assertEqual(dual.objective.value, 31)

    def test_variable_with_upper_bound(self):
        self.model.variables["x"].ub = 4
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 7)
        self.assertEqual(dual.objective.value, 7)

    def test_with_minimization(self):
        self.model.objective = Objective(self.model.objective.expression * -1, direction="min")
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, -31)
        self.assertEqual(dual.objective.value, -31)

    def test_equality_constraint_not_zero(self):
        c2 = self.model.constraints[1]
        c2.ub = 1
        c2.lb = 1
        dual = convert_linear_problem_to_dual(self.model, maintain_standard_form=False)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 32)
        self.assertEqual(dual.objective.value, 32)

    def test_empty_constraint(self):
        c4 = Constraint(0, lb=0, ub=13)
        self.model.add(c4)
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 31)
        self.assertEqual(dual.objective.value, 31)

    def test_free_constraint(self):
        c4 = Constraint(self.model.variables["x"], lb=None, ub=None)
        self.model.add(c4)
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 31)
        self.assertEqual(dual.objective.value, 31)

    def test_zero_bound_variable(self):
        w = Variable("w", lb=0, ub=0)
        self.model.add(w)
        dual = convert_linear_problem_to_dual(self.model)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 31)
        self.assertEqual(dual.objective.value, 31)

    def test_explicit_model(self):
        dual = Model()
        convert_linear_problem_to_dual(self.model, dual_model=dual)
        primal_status = self.model.optimize()
        dual_status = dual.optimize()

        self.assertEqual(primal_status, optlang.interface.OPTIMAL)
        self.assertEqual(dual_status, optlang.interface.OPTIMAL)

        self.assertEqual(self.model.objective.value, 31)
        self.assertEqual(dual.objective.value, 31)



import unittest
import optlang


class SymbolicsTestCase(unittest.TestCase):

    def test_add_identity(self):
        self.assertEqual(optlang.symbolics.add(), 0.0)

    def test_mul_identity(self):
        self.assertEqual(optlang.symbolics.mul(), 1.0)



# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

from optlang.interface import Constraint, Variable, Objective
from optlang.expression_parsing import parse_optimization_expression

import unittest


def _quad_terms_to_expression(terms):
    args = []
    for term, coef in terms.items():
        term = list(term)
        args.append(coef * term[0] * term[-1])
    return sum(args)


def _compare_term_dicts(test_case, dict1, dict2):
    for term, coef1 in dict1.items():
        coef2 = dict2[term]
        test_case.assertEqual(coef1 - coef2, 0)


class ExpressionParsingTestCase(unittest.TestCase):
    def setUp(self):
        self.vars = [Variable(name) for name in ["x", "y", "z", "v", "w"]]

    def test_parse_linear_expression(self):
        x, y, z = self.vars[:3]
        offset = 3
        expr = 1 * x + 2 * y - 3 * z + offset
        target = {x: 1, y: 2, z: -3}

        offset_const, linear_terms_const, quad_terms_const = parse_optimization_expression(Constraint(expr, lb=0))
        offset_obj, linear_terms_obj, quad_terms_obj = parse_optimization_expression(Objective(expr), linear=False)

        self.assertEqual(offset_const, 0)
        self.assertEqual(offset_obj, offset)
        _compare_term_dicts(self, linear_terms_const, target)
        _compare_term_dicts(self, linear_terms_obj, target)
        self.assertEqual(quad_terms_const, {})
        self.assertEqual(quad_terms_obj, {})

    def test_parse_quadratic_expression(self):
        x, y, z = self.vars[:3]

        offset = 4
        expr = 2 * x**2 + 3 * x * y - 4 * z**2 + offset
        target = {frozenset([x]): 2, frozenset([z]): -4, frozenset([x, y]): 3}

        offset_const, linear_terms_const, quad_terms_const = parse_optimization_expression(Constraint(expr, lb=0), linear=False)
        offset_obj, linear_terms_obj, quad_terms_obj = parse_optimization_expression(Objective(expr), quadratic=True)

        self.assertEqual(offset_const, 0)
        self.assertEqual(offset_obj, offset)
        self.assertEqual(linear_terms_const, {})
        self.assertEqual(linear_terms_obj, {})
        _compare_term_dicts(self, quad_terms_const, target)
        _compare_term_dicts(self, quad_terms_obj, target)
        self.assertEqual((_quad_terms_to_expression(quad_terms_obj) - (expr - offset)).expand(), 0)
        self.assertEqual((_quad_terms_to_expression(quad_terms_const) - (expr - offset)).expand(), 0)

    def test_parse_non_expanded_quadratic_expression(self):
        x, y, z = self.vars[:3]

        offset = 5
        expr = (x + y)**2 - (z - 2)**2 + offset
        target = {frozenset([x]): 1, frozenset([y]): 1, frozenset([x, y]): 2, frozenset([z]): -1}
        linear_target = {z: 4}

        constraint = Constraint(expr, lb=0)
        offset_const, linear_terms_const, quad_terms_const = parse_optimization_expression(
            constraint,
            quadratic=True
        )
        offset_obj, linear_terms_obj, quad_terms_obj = parse_optimization_expression(
            Objective(expr),
            expression=expr,
            linear=False
        )

        self.assertEqual(offset_const - constraint.lb, -4 + offset)
        self.assertEqual(offset_obj, -4 + offset)
        _compare_term_dicts(self, linear_terms_const, linear_target)
        _compare_term_dicts(self, linear_terms_obj, linear_target)
        _compare_term_dicts(self, quad_terms_const, target)
        _compare_term_dicts(self, quad_terms_obj, target)



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Solver interface for the OSQP solver.

Wraps the OSQP solver by subclassing and extending :class:`Model`,
:class:`Variable`, and :class:`Constraint` from :mod:`interface`.

To use this interface, install the OSQP solver and the bundled python
interface.
Make sure that 'import osqp' runs without error.
"""
import logging
import six
import pickle
import numpy as np
from numpy import array, concatenate, Infinity, zeros, isnan

from optlang import interface, symbolics, available_solvers
from optlang.util import inheritdocstring
from optlang.expression_parsing import parse_optimization_expression
from optlang.exceptions import SolverError

from scipy.sparse import csc_matrix

from optlang.symbolics import add, mul

log = logging.getLogger(__name__)

try:
    import osqp as osqp
except ImportError:
    try:
        import cuosqp as osqp

        log.warning(
            "cuOSQP is still experimental and may not work for all problems. "
            "It may not converge as fast as the normal osqp package or not "
            "converge at all."
        )
    except ImportError:
        raise ImportError("The osqp_interface requires osqp or cuosqp!")


_STATUS_MAP = {
    "interrupted by user": interface.ABORTED,
    "run time limit reached": interface.TIME_LIMIT,
    "feasible": interface.FEASIBLE,
    "primal infeasible": interface.INFEASIBLE,
    "dual infeasible": interface.INFEASIBLE,
    "primal infeasible inaccurate": interface.INFEASIBLE,
    "dual infeasible inaccurate": interface.INFEASIBLE,
    "solved inaccurate": interface.NUMERIC,
    "solved": interface.OPTIMAL,
    "maximum iterations reached": interface.ITERATION_LIMIT,
    "unsolved": interface.SPECIAL,
    "problem non convex": interface.SPECIAL,
    "non-existing-status": "Here for testing that missing statuses are handled.",
}

_LP_METHODS = ("auto", "primal")

_QP_METHODS = ("auto", "primal")

_TYPES = ("continuous",)


class OSQPProblem(object):
    """A concise representation of an OSQP problem.

    OSQP assumes that the problem will be pretty much immutable. This is
    a small intermediate layer based on dictionaries that is fast to modify
    but can also be converted to an OSQP problem without too much hassle.
    """

    def __init__(self):
        self.variables = set()
        self.constraints = set()
        self.constraint_coefs = dict()
        self.constraint_lbs = dict()
        self.constraint_ubs = dict()
        self.variable_lbs = dict()
        self.variable_ubs = dict()
        self.obj_linear_coefs = dict()
        self.obj_quadratic_coefs = dict()
        self.primals = {}
        self.cprimals = {}
        self.duals = {}
        self.vduals = {}
        self.obj_value = None
        self.direction = -1
        self.info = None
        self.status = None
        self.settings = {
            "linsys_solver": "qdldl",
            "max_iter": 100000,
            "eps_abs": 1e-6,
            "eps_rel": 1e-6,
            "eps_prim_inf": 1e-6,
            "eps_dual_inf": 1e-6,
            "polish": True,
            "verbose": False,
            "scaling": 0,
            "time_limit": 0,
            "adaptive_rho": True,
            "rho": 1.0,
            "alpha": 1.6,
        }
        self.__solution = None

    def build(self):
        """Build the problem instance."""
        d = self.direction
        vmap = dict(zip(self.variables, range(len(self.variables))))
        nv = len(self.variables)
        cmap = dict(zip(self.constraints, range(len(self.constraints))))
        nc = len(self.constraints)
        if len(self.obj_quadratic_coefs) > 0:
            P = array(
                [
                    [vmap[vn[0]], vmap[vn[1]], coef * d * 2.0]
                    for vn, coef in six.iteritems(self.obj_quadratic_coefs)
                ]
            )
            P = csc_matrix(
                (P[:, 2], (P[:, 0].astype("int64"), P[:, 1].astype("int64"))),
                shape=(nv, nv),
            )
        else:
            P = None
        q = zeros(nv)
        q[[vmap[vn] for vn in self.obj_linear_coefs]] = list(
            self.obj_linear_coefs.values()
        )
        q = q * d
        Av = array([[vmap[k] + nc, vmap[k], 1.0] for k in self.variables])
        vbounds = array(
            [[self.variable_lbs[vn], self.variable_ubs[vn]] for vn in self.variables]
        )
        if len(self.constraint_coefs) > 0:
            A = array(
                [
                    [cmap[vn[0]], vmap[vn[1]], coef]
                    for vn, coef in six.iteritems(self.constraint_coefs)
                ]
            )
            bounds = array(
                [
                    [self.constraint_lbs[cn], self.constraint_ubs[cn]]
                    for cn in self.constraints
                ]
            )
            A = concatenate((A, Av))
            bounds = concatenate((bounds, vbounds))
        else:
            A = Av
            bounds = vbounds
        if A.shape[0] == 0:
            A = None
        else:
            A = csc_matrix(
                (A[:, 2], (A[:, 0].astype("int64"), A[:, 1].astype("int64"))),
                shape=(nc + nv, nv),
            )
        return P, q, A, bounds

    def solve(self):
        """Solve the OSQP problem."""
        settings = self.settings.copy()
        P, q, A, bounds = self.build()
        solver = osqp.OSQP()
        if P is None:
            # see https://github.com/cvxgrp/cvxpy/issues/898
            settings.update({"adaptive_rho": 0, "rho": 1.0, "alpha": 1.0})
        solver.setup(P=P, q=q, A=A, l=bounds[:, 0], u=bounds[:, 1], **settings)  # noqa
        if self.__solution is not None:
            if self.still_valid(A, bounds):
                solver.warm_start(x=self.__solution["x"], y=self.__solution["y"])
                solver.update_settings(rho=self.__solution["rho"])
        solution = solver.solve()
        nc = len(self.constraints)
        nv = len(self.variables)
        if not solution.x[0] is None:
            self.primals = dict(zip(self.variables, solution.x))
            self.vduals = dict(zip(self.variables, solution.y[nc : (nc + nv)]))
            if nc > 0:
                self.cprimals = dict(zip(self.constraints, A.dot(solution.x)[0:nc]))
                self.duals = dict(zip(self.constraints, solution.y[0:nc]))
        if not isnan(solution.info.obj_val):
            self.obj_value = solution.info.obj_val * self.direction
            self.status = solution.info.status
        else:
            self.status = "primal infeasible"
            self.obj_value = None
        self.info = solution.info
        self.__solution = {
            "x": solution.x,
            "y": solution.y,
            "rho": solution.info.rho_estimate,
        }

    def reset(self, everything=False):
        """Reset the public solver solution."""
        self.info = None
        self.primals = {}
        self.cprimals = {}
        self.duals = {}
        self.vduals = {}
        if everything:
            self.__solution = None

    def still_valid(self, A, bounds):
        """Check if previous solutions is still feasible."""
        if len(self.__solution["x"]) != len(self.variables) or len(
            self.__solution["y"]
        ) != len(self.constraints):
            return False
        c = A.dot(self.__solution["x"])
        ea = self.settings["eps_abs"]
        er = self.settings["eps_rel"]
        valid = np.all(
            (c + er * np.abs(c) + ea >= bounds[:, 0])
            & (c - er * np.abs(c) - ea <= bounds[:, 1])
        )
        return valid

    def clean(self):
        """Remove unused variables and constraints."""
        self.reset()
        self.variable_lbs = {
            k: v for k, v in six.iteritems(self.variable_lbs) if k in self.variables
        }
        self.variable_ubs = {
            k: v for k, v in six.iteritems(self.variable_ubs) if k in self.variables
        }
        self.constraint_coefs = {
            k: v
            for k, v in six.iteritems(self.constraint_coefs)
            if k[0] in self.constraints and k[1] in self.variables
        }
        self.obj_linear_coefs = {
            k: v for k, v in six.iteritems(self.obj_linear_coefs) if k in self.variables
        }
        self.obj_quadratic_coefs = {
            k: v
            for k, v in six.iteritems(self.obj_quadratic_coefs)
            if k[0] in self.variables and k[1] in self.variables
        }
        self.constraint_lbs = {
            k: v for k, v in six.iteritems(self.constraint_lbs) if k in self.constraints
        }
        self.constraint_ubs = {
            k: v for k, v in six.iteritems(self.constraint_ubs) if k in self.constraints
        }

    def rename_constraint(self, old, new):
        """Rename a constraint."""
        self.reset()
        self.constraints.remove(old)
        self.constraints.add(new)
        name_map = {k: k for k in self.constraints}
        name_map[old] = new
        self.constraint_coefs = {
            (name_map[k[0]], k[1]): v for k, v in six.iteritems(self.constraint_coefs)
        }
        self.constraint_lbs[new] = self.constraint_lbs.pop(old)
        self.constraint_ubs[new] = self.constraint_ubs.pop(old)

    def rename_variable(self, old, new):
        """Rename a variable."""
        self.reset()
        self.variables.remove(old)
        self.variables.add(new)
        name_map = {k: k for k in self.variables}
        name_map[old] = new
        self.constraint_coefs = {
            (k[0], name_map[k[1]]): v for k, v in six.iteritems(self.constraint_coefs)
        }
        self.obj_quadratic_coefs = {
            (name_map[k[0]], name_map[k[1]]): v
            for k, v in six.iteritems(self.obj_quadratic_coefs)
        }
        self.obj_linear_coefs = {
            name_map[k]: v for k, v in six.iteritems(self.obj_linear_coefs)
        }
        self.variable_lbs[new] = self.variable_lbs.pop(old)
        self.variable_ubs[new] = self.variable_ubs.pop(old)


@six.add_metaclass(inheritdocstring)
class Variable(interface.Variable):
    def __init__(self, name, *args, **kwargs):
        super(Variable, self).__init__(name, **kwargs)

    @interface.Variable.type.setter
    def type(self, value):
        if self.problem is not None:
            if value not in _TYPES:
                raise ValueError(
                    "OSQP cannot handle variables of type '%s'. " % value
                    + "The following variable types are available: "
                    + ", ".join(_TYPES)
                )
        super(Variable, Variable).type.fset(self, value)

    def _get_primal(self):
        return self.problem.problem.primals.get(self.name, None)

    @property
    def dual(self):
        if self.problem is not None:
            return self.problem.problem.vduals.get(self.name, None)
        return None

    @interface.Variable.name.setter
    def name(self, value):
        old_name = getattr(self, "name", None)
        super(Variable, Variable).name.fset(self, value)
        if getattr(self, "problem", None) is not None:
            if old_name != value:
                self.problem.problem.rename_variable(old_name, value)


@six.add_metaclass(inheritdocstring)
class Constraint(interface.Constraint):
    _INDICATOR_CONSTRAINT_SUPPORT = False

    def __init__(self, expression, sloppy=False, *args, **kwargs):
        super(Constraint, self).__init__(expression, *args, sloppy=sloppy, **kwargs)

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            self.problem.problem.reset()
            for var, coef in six.iteritems(coefficients):
                self.problem.problem.constraint_coefs[(self.name, var.name)] = float(
                    coef
                )
        else:
            raise Exception(
                "Can't change coefficients if constraint is not associated "
                "with a model."
            )

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            coefs = {
                v: self.problem.problem.constraint_coefs.get((self.name, v.name), 0.0)
                for v in variables
            }
            return coefs
        else:
            raise Exception(
                "Can't get coefficients from solver if constraint is not " "in a model"
            )

    def _get_expression(self):
        if self.problem is not None:
            variables = self.problem._variables
            all_coefs = self.problem.problem.constraint_coefs
            coefs = [(v, all_coefs.get((self.name, v.name), 0.0)) for v in variables]
            expression = add([mul((symbolics.Real(co), v)) for (v, co) in coefs])
            self._expression = expression
        return self._expression

    @property
    def problem(self):
        return self._problem

    @problem.setter
    def problem(self, value):
        if value is None:
            # Update expression from solver instance one last time
            self._get_expression()
        self._problem = value

    @property
    def primal(self):
        if self.problem is None:
            return None
        return self.problem.problem.cprimals.get(self.name, None)

    @property
    def dual(self):
        if self.problem is None:
            return None
        d = self.problem.problem.duals.get(self.name, None)
        if d is not None:
            d = -d
        return d

    @interface.Constraint.name.setter
    def name(self, value):
        old_name = self.name
        super(Constraint, Constraint).name.fset(self, value)
        if getattr(self, "problem", None) is not None:
            if old_name != value:
                self.problem.problem.rename_constraint(old_name, value)

    @interface.Constraint.lb.setter
    def lb(self, value):
        self._check_valid_lower_bound(value)
        if getattr(self, "problem", None) is not None:
            lb = -Infinity if value is None else float(value)
            self.problem.problem.constraint_lbs[self.name] = lb
        self._lb = value

    @interface.Constraint.ub.setter
    def ub(self, value):
        self._check_valid_upper_bound(value)
        if getattr(self, "problem", None) is not None:
            ub = Infinity if value is None else float(value)
            self.problem.problem.constraint_ubs[self.name] = ub
        self._ub = value

    def __iadd__(self, other):
        # if self.problem is not None:
        #     self.problem._add_to_constraint(self.index, other)
        if self.problem is not None:
            problem_reference = self.problem
            self.problem._remove_constraint(self)
            super(Constraint, self).__iadd__(other)
            problem_reference._add_constraint(self, sloppy=False)
        else:
            super(Constraint, self).__iadd__(other)
        return self


@six.add_metaclass(inheritdocstring)
class Objective(interface.Objective):
    def __init__(self, expression, sloppy=False, **kwargs):
        super(Objective, self).__init__(expression, sloppy=sloppy, **kwargs)
        self._expression_expired = False
        if not (sloppy or self.is_Linear or self.is_Quadratic):
            raise ValueError("OSQP only supports linear and quadratic objectives.")

    @property
    def value(self):
        if getattr(self, "problem", None) is None:
            return None
        if self.problem.problem.obj_value is None:
            return None
        return self.problem.problem.obj_value + self.problem._objective_offset

    @interface.Objective.direction.setter
    def direction(self, value):
        super(Objective, Objective).direction.__set__(self, value)
        if value == "min":
            self.problem.problem.direction = 1
        else:
            self.problem.problem.direction = -1

    def _get_expression(self):
        if (
            self.problem is not None
            and self._expression_expired
            and len(self.problem._variables) > 0
        ):
            model = self.problem
            vars = model._variables
            expression = add(
                [
                    coef * vars[vn]
                    for vn, coef in six.iteritems(model.problem.obj_linear_coefs)
                ]
            )
            q_ex = add(
                [
                    coef * vars[vn[0]] * vars[vn[1]]
                    for vn, coef in six.iteritems(model.problem.obj_quadratic_coefs)
                ]
            )
            expression += q_ex
            self._expression = expression
            self._expression_expired = False
        return self._expression

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            for v, coef in six.iteritems(coefficients):
                self.problem.problem.obj_linear_coefs[v.name] = float(coef)
            self._expression_expired = True
        else:
            raise Exception(
                "Can't change coefficients if objective is not associated "
                "with a model."
            )

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            coefs = {
                v: self.problem.problem.obj_linear_coefs.get(v.name, 0.0)
                for v in variables
            }
            return coefs
        else:
            raise Exception(
                "Can't get coefficients from solver if objective " "is not in a model"
            )


@six.add_metaclass(inheritdocstring)
class Configuration(interface.MathematicalProgrammingConfiguration):
    def __init__(
        self,
        lp_method="primal",
        presolve=False,
        verbosity=0,
        timeout=None,
        qp_method="primal",
        linear_solver="qdldl",
        *args,
        **kwargs
    ):
        super(Configuration, self).__init__(*args, **kwargs)
        self.lp_method = lp_method
        self.presolve = presolve
        self.verbosity = verbosity
        self.timeout = timeout
        self.qp_method = qp_method
        self.linear_solver = linear_solver
        if "tolerances" in kwargs:
            for key, val in six.iteritems(kwargs["tolerances"]):
                if key in self._tolerance_functions():
                    setattr(self.tolerances, key, val)

    @property
    def lp_method(self):
        """The algorithm used to solve LP problems."""
        return "primal"

    @lp_method.setter
    def lp_method(self, lp_method):
        if lp_method not in _LP_METHODS:
            raise ValueError(
                "LP Method %s is not valid (choose one of: %s)"
                % (lp_method, ", ".join(_LP_METHODS))
            )

    @property
    def linear_solver(self):
        return self._linear_solver

    @linear_solver.setter
    def linear_solver(self, solver):
        if solver not in ("qdldl", "mkl pardiso"):
            raise ValueError(
                "%s is not valid (choose either `qdldl` or `mkl pardiso`)" % solver
            )
        if getattr(self, "problem", None) is not None:
            self.problem.problem.settings["linsys_solver"] = solver
        self._linear_solver = solver

    def _set_presolve(self, value):
        if getattr(self, "problem", None) is not None:
            if value is True:
                self.problem.problem.settings["scaling"] = 10
            elif value is False or value == "auto":
                self.problem.problem.settings["scaling"] = 0
            else:
                raise ValueError(
                    str(value) + " is not a valid presolve parameter. Must be True, "
                    "False or 'auto'."
                )

    @property
    def presolve(self):
        return self.problem.problem.settings["scaling"] > 0

    @presolve.setter
    def presolve(self, value):
        self._set_presolve(value)
        self._presolve = value

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        if getattr(self, "problem", None) is not None:
            self.problem.problem.settings["verbose"] = int(value > 0)
        self._verbosity = value

    @property
    def timeout(self):
        return self.problem.problem.settings["time_limit"]

    @timeout.setter
    def timeout(self, value):
        if getattr(self, "problem", None) is not None:
            if value is None:
                self.problem.problem.settings["time_limit"] = 0
            else:
                self.problem.problem.settings["time_limit"] = value

    def __getstate__(self):
        return {
            "presolve": self.presolve,
            "timeout": self.timeout,
            "verbosity": self.verbosity,
            "linear_solver": self.linear_solver,
            "tolerances": {
                "feasibility": self.tolerances.feasibility,
                "optimality": self.tolerances.optimality,
                "integrality": self.tolerances.integrality,
            },
        }

    def __setstate__(self, state):
        for key, val in six.iteritems(state):
            if key != "tolerances":
                setattr(self, key, val)
        for key, val in six.iteritems(state["tolerances"]):
            if key in self._tolerance_functions():
                setattr(self.tolerances, key, val)

    @property
    def qp_method(self):
        """Change the algorithm used to optimize QP problems."""
        return "primal"

    @qp_method.setter
    def qp_method(self, value):
        if value not in _QP_METHODS:
            raise ValueError(
                "%s is not a valid qp_method. Choose between %s"
                % (value, str(_QP_METHODS))
            )
        self._qp_method = value

    def _get_feasibility(self):
        return self.problem.problem.settings["eps_prim_inf"]

    def _set_feasibility(self, value):
        self.problem.problem.settings["eps_prim_inf"] = value
        self.problem.problem.settings["eps_dual_inf"] = value

    def _get_integrality(self):
        return None

    def _set_integrality(self, value):
        pass

    def _get_optimality(self):
        return self.problem.problem.settings["eps_abs"]

    def _set_optimality(self, value):
        self.problem.problem.settings["eps_abs"] = value
        self.problem.problem.settings["eps_rel"] = value

    def _tolerance_functions(self):
        return {
            "feasibility": (self._get_feasibility, self._set_feasibility),
            "optimality": (self._get_optimality, self._set_optimality),
            "integrality": (self._get_integrality, self._set_integrality),
        }


@six.add_metaclass(inheritdocstring)
class Model(interface.Model):
    def _initialize_problem(self):
        self.problem = OSQPProblem()

    def _initialize_model_from_problem(self, problem, vc_mapping=None, offset=0):
        if not isinstance(problem, OSQPProblem):
            raise TypeError("Provided problem is not a valid OSQP model.")
        self.problem = problem
        for name in self.problem.variables:
            var = Variable(
                name,
                lb=self.problem.variable_lbs[name],
                ub=self.problem.variable_ubs[name],
                problem=self,
            )
            super(Model, self)._add_variables([var])

        for name in self.problem.constraints:
            # Since constraint expressions are lazily retrieved from the
            # solver they don't have to be built here
            lhs = symbolics.Integer(0)
            constr = Constraint(
                lhs,
                lb=self.problem.constraint_lbs[name],
                ub=self.problem.constraint_ubs[name],
                name=name,
                problem=self,
            )

            super(Model, self)._add_constraints([constr], sloppy=True)

        if vc_mapping is None:
            for constr in self.constraints:
                name = constr.name
                for variable in constr.variables:
                    try:
                        self._variables_to_constraints_mapping[variable.name].add(name)
                    except KeyError:
                        self._variables_to_constraints_mapping[variable.name] = set(
                            [name]
                        )
        else:
            self._variables_to_constraints_mapping = vc_mapping

        linear_expression = add(
            [
                coef * self._variables[vn]
                for vn, coef in six.iteritems(self.problem.obj_linear_coefs)
            ]
        )
        quadratic_expression = add(
            [
                coef * self._variables[vn[0]] * self._variables[vn[1]]
                for vn, coef in six.iteritems(self.problem.obj_quadratic_coefs)
            ]
        )

        self._objective_offset = offset
        self._objective = Objective(
            linear_expression + quadratic_expression + offset,
            problem=self,
            direction={-1: "max", 1: "min"}[self.problem.direction],
            name="osqp_objective",
        )

    @property
    def objective(self):
        return self._objective

    @objective.setter
    def objective(self, value):
        value.problem = None
        if self._objective is not None:  # Reset previous objective
            self.problem.obj_linear_coefs = dict()
            self.problem.obj_quadratic_coefs = dict()
        super(Model, self.__class__).objective.fset(self, value)
        self.update()
        expression = self._objective._expression
        (
            offset,
            linear_coefficients,
            quadratic_coeffients,
        ) = parse_optimization_expression(value, quadratic=True, expression=expression)
        self._objective_offset = offset
        if linear_coefficients:
            self.problem.obj_linear_coefs = {
                v.name: float(c) for v, c in six.iteritems(linear_coefficients)
            }

        for key, coef in six.iteritems(quadratic_coeffients):
            if len(key) == 1:
                var = six.next(iter(key))
                self.problem.obj_quadratic_coefs[(var.name, var.name)] = float(coef)
            else:
                var1, var2 = key
                self.problem.obj_quadratic_coefs[(var1.name, var2.name)] = 0.5 * float(
                    coef
                )
                self.problem.obj_quadratic_coefs[(var2.name, var1.name)] = 0.5 * float(
                    coef
                )

        self._set_objective_direction(value.direction)
        value.problem = self

    def _set_objective_direction(self, direction):
        self.problem.direction = {"min": 1, "max": -1}[direction]

    def _get_primal_values(self):
        if len(self.problem.primals) == 0:
            raise SolverError("The problem has not been solved yet!")
        primal_values = [self.problem.primals[v.name] for v in self._variables]
        return primal_values

    def _get_reduced_costs(self):
        if len(self.problem.duals) == 0:
            raise SolverError("The problem has not been solved yet!")
        reduced_costs = [self.problem.vduals[v.name] for v in self._variables]
        return reduced_costs

    def _get_constraint_values(self):
        if len(self.problem.primals) == 0:
            raise SolverError("The problem has not been solved yet!")
        constraint_primals = [self.problem.cprimals[c.name] for c in self._constraints]
        return constraint_primals

    def _get_shadow_prices(self):
        if len(self.problem.primals) == 0:
            raise SolverError("The problem has not been solved yet!")
        dual_values = [-self.problem.duals[c.name] for c in self._constraints]
        return dual_values

    @property
    def is_integer(self):
        return False

    def _optimize(self):
        self.update()
        self.problem.solve()
        osqp_status = self.problem.status
        self._original_status = osqp_status
        status = _STATUS_MAP[osqp_status]
        return status

    def _set_variable_bounds_on_problem(self, var_lb, var_ub):
        self.problem.reset()
        for var, val in var_lb:
            lb = -Infinity if val is None else float(val)
            self.problem.variable_lbs[var.name] = lb
        for var, val in var_ub:
            ub = Infinity if val is None else val
            self.problem.variable_ubs[var.name] = float(ub)

    def _add_variables(self, variables):
        super(Model, self)._add_variables(variables)
        self.problem.reset()
        for variable in variables:
            lb = -Infinity if variable.lb is None else float(variable.lb)
            ub = Infinity if variable.ub is None else float(variable.ub)
            self.problem.variables.add(variable.name)
            self.problem.variable_lbs[variable.name] = lb
            self.problem.variable_ubs[variable.name] = ub
            variable.problem = self

    def _remove_variables(self, variables):
        # Not calling parent method to avoid expensive variable removal from sympy expressions
        if self.objective is not None:
            self.objective._expression = self.objective.expression.xreplace(
                {var: 0 for var in variables}
            )
        for variable in variables:
            del self._variables_to_constraints_mapping[variable.name]
            variable.problem = None
            del self._variables[variable.name]
            self.problem.variables.remove(variable.name)
        self.problem.clean()

    def _add_constraints(self, constraints, sloppy=False):
        super(Model, self)._add_constraints(constraints, sloppy=sloppy)
        self.problem.reset()
        for constraint in constraints:
            constraint._problem = None  # This needs to be done in order to not trigger constraint._get_expression()
            if constraint.is_Linear:
                _, coeff_dict, _ = parse_optimization_expression(constraint)
                lb = -Infinity if constraint.lb is None else float(constraint.lb)
                ub = Infinity if constraint.ub is None else float(constraint.ub)
                self.problem.constraints.add(constraint.name)
                self.problem.constraint_coefs.update(
                    {
                        (constraint.name, v.name): float(co)
                        for v, co in six.iteritems(coeff_dict)
                    }
                )
                self.problem.constraint_lbs[constraint.name] = lb
                self.problem.constraint_ubs[constraint.name] = ub
                constraint.problem = self
            elif constraint.is_Quadratic:
                raise NotImplementedError(
                    "Quadratic constraints (like %s) are not supported "
                    "in OSQP yet." % constraint
                )
            else:
                raise ValueError(
                    "OSQP only supports linear or quadratic constraints. "
                    "%s is neither linear nor quadratic." % constraint
                )

    def _remove_constraints(self, constraints):
        super(Model, self)._remove_constraints(constraints)
        for constraint in constraints:
            self.problem.constraints.remove(constraint.name)
        self.problem.clean()

    def _get_variable_indices(self, names):
        vmap = dict(zip(self.variables, range(len(self.variables))))
        return [vmap[n] for n in names]

    def __setstate__(self, d):
        self.__init__()
        osqp = pickle.loads(d["osqp"])
        # workaround to conserve the order
        osqp.variables = d["variables"]
        osqp.constraints = d["constraints"]
        self._initialize_model_from_problem(
            osqp, vc_mapping=d["v_to_c"], offset=d["offset"]
        )
        osqp.variables = set(osqp.variables)
        osqp.constraints = set(osqp.constraints)
        self.configuration = Configuration()
        self.configuration.problem = self
        self.configuration.__setstate__(d["config"])

    def __getstate__(self):
        self.problem.reset()
        self.update()
        return {
            "osqp": pickle.dumps(self.problem),
            "variables": [v.name for v in self._variables],
            "constraints": [c.name for c in self._constraints],
            "v_to_c": self._variables_to_constraints_mapping,
            "config": self.configuration.__getstate__(),
            "offset": getattr(self, "_objective_offset", 0.0),
        }

    @classmethod
    def from_lp(self, lp_problem_str):
        """Read a model from an LP file.

        OSQP does not have an integrated LP reader so it will either use
        cplex or glpk to read the model. This means that QP problems will
        currently require cplex to be read :(
        """
        if available_solvers["CPLEX"]:
            from optlang import cplex_interface

            mod = cplex_interface.Model.from_lp(lp_problem_str)
            mod.configuration.lp_method = "auto"
            mod.configuration.qp_method = "auto"
            return super(Model, self).clone(mod)
        else:
            from optlang import glpk_interface

            mod = glpk_interface.Model.from_lp(lp_problem_str)
            mod.configuration.lp_method = "auto"
            return super(Model, self).clone(mod)



# Copyright 2016 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import abc
import unittest

import six
from optlang import interface
from optlang import symbolics
import optlang
import pickle
import json
import copy
import os
import sympy
from functools import partial

__test__ = False

TESTMODELPATH = os.path.join(os.path.dirname(__file__), 'data/model.json')
TESTMILPMODELPATH = os.path.join(os.path.dirname(__file__), 'data/simple_milp.json')


@six.add_metaclass(abc.ABCMeta)
class AbstractVariableTestCase(unittest.TestCase):
    @unittest.skip('Abstract test method not implemented.')
    def test_magic(self):
        pass

    def setUp(self):
        self.var = self.interface.Variable('test')
        self.model = self.interface.Model()

    def test_set_wrong_type_raises(self):
        self.assertRaises(ValueError, self.interface.Variable, name="test", type="mayo")
        self.assertRaises(Exception, setattr, self.var, 'type', 'ketchup')
        self.model.add(self.var)
        self.model.update()
        self.assertRaises(ValueError, setattr, self.var, "type", "mustard")
        self.var.type = "integer"
        self.assertEqual(self.var.type, "integer")

    def test_change_type(self):
        var = self.interface.Variable("test")
        var.type = "binary"
        self.assertEqual(var.type, "binary")

    def test_change_name(self):
        self.model.add(self.var)
        self.model.update()
        self.var.name = "test_2"
        self.assertEqual(self.var.name, "test_2")
        self.assertEqual(str(self.var), self.var.name)
        self.model.remove(self.var)
        self.model.update()

    def test_non_string_name_raises(self):
        for name in [2, None, True, ["name1", "name2"]]:
            with self.assertRaises(TypeError):
                self.interface.Variable(name)

    @abc.abstractmethod
    def test_get_primal(self):
        pass

    def test_get_dual(self):
        with open(TESTMODELPATH) as infile:
            model = self.interface.Model.from_json(json.load(infile))
        model.optimize()
        self.assertEqual(model.status, 'optimal')
        self.assertAlmostEqual(model.objective.value, 0.8739215069684305)
        self.assertTrue(isinstance(model.variables[0].dual, float))

    def test_setting_lower_bound_higher_than_upper_bound_raises(self):
        self.model.add(self.var)
        self.var.ub = 0
        self.assertRaises(ValueError, setattr, self.model.variables["test"], 'lb', 100.)

    def test_setting_nonnumerical_bounds_raises(self):
        self.assertRaises(TypeError, setattr, self.var, "lb", "Minestrone")
        self.assertRaises(TypeError, setattr, self.var, "ub", "Minestrone")
        self.model.add(self.var)
        self.assertRaises(TypeError, setattr, self.model.variables[0], 'lb', 'Chicken soup')
        self.assertRaises(TypeError, setattr, self.model.variables[0], 'ub', 'Chicken soup')

    @abc.abstractmethod
    def test_changing_variable_names_is_reflected_in_the_solver(self):
        pass

    def test_setting_bounds(self):
        self.var.ub = 5
        self.model.objective = self.interface.Objective(self.var)
        self.model.optimize()
        self.assertEqual(self.var.primal, 5)
        self.var.ub = 4
        self.model.optimize()
        self.assertEqual(self.var.primal, 4)
        self.var.lb = -3
        self.model.objective.direction = "min"
        self.model.optimize()
        self.assertEqual(self.var.primal, -3)
        self.var.lb = sympy.Number(-4)  # Sympy numbers should be valid bounds
        self.model.optimize()
        self.assertEqual(self.var.primal, -4)

    def test_set_bounds_method(self):
        var = self.interface.Variable("test", lb=-10)
        c = self.interface.Constraint(var, lb=-100)
        model = self.interface.Model()
        obj = self.interface.Objective(var)
        model.add(c)
        model.objective = obj

        for lb, ub in ((1, 10), (-1, 5), (11, 12)):
            obj.direction = "max"
            var.set_bounds(lb, ub)
            model.optimize()
            self.assertAlmostEqual(var.primal, ub)
            obj.direction = "min"
            model.optimize()
            self.assertAlmostEqual(var.primal, lb)

        var.set_bounds(None, 0)
        model.optimize()
        self.assertAlmostEqual(var.primal, -100)

        obj.direction = "max"
        var.set_bounds(1, None)
        self.assertEqual(model.optimize(), optlang.interface.UNBOUNDED)

        self.assertRaises(ValueError, var.set_bounds, 2, 1)

    def test_set_bounds_to_none(self):
        model = self.model
        var = self.var
        model.objective = self.interface.Objective(var)
        self.assertEqual(model.optimize(), interface.UNBOUNDED)
        var.ub = 10
        self.assertEqual(model.optimize(), interface.OPTIMAL)
        var.ub = None
        self.assertEqual(model.optimize(), interface.UNBOUNDED)
        self.model.objective.direction = "min"
        var.lb = -10
        self.assertEqual(model.optimize(), interface.OPTIMAL)
        var.lb = None
        self.assertEqual(model.optimize(), interface.UNBOUNDED)

    def test_invalid_name_raises(self):
        with self.assertRaises(Exception):
            self.interface.Variable("")
        with self.assertRaises(Exception):
            self.interface.Variable("This space")
        with self.assertRaises(Exception):
            self.interface.Variable("This\ttab")

    def test_new_invalid_name_raises(self):
        with self.assertRaises(Exception):
            self.var.name = ""
        with self.assertRaises(Exception):
            self.var.name = "This space"
        with self.assertRaises(Exception):
            self.var.name = "This\ttab"

        self.assertEqual(self.var.name, 'test')


@six.add_metaclass(abc.ABCMeta)
class AbstractConstraintTestCase(unittest.TestCase):

    def setUp(self):
        with open(TESTMODELPATH) as infile:
            self.model = self.interface.Model.from_json(json.load(infile))
        self.constraint = self.interface.Constraint(
            self.interface.Variable('chip') + self.interface.Variable('chap'),
            name='woodchips',
            lb=100
        )

    def test_indicator_constraint_support(self):
        if self.interface.Constraint._INDICATOR_CONSTRAINT_SUPPORT:
            constraint = self.interface.Constraint(
                self.interface.Variable('chip_2'),
                indicator_variable=self.interface.Variable('chip', type='binary'), active_when=0, lb=0,
                ub=0,
                name='indicator_constraint_fwd_1'
            )
            model = self.interface.Model()
            model.add(constraint)
            model.update()
            self.assertEqual(constraint.problem, model)
            self.assertIn(constraint, model.constraints)
            model.remove(constraint.name)
            model.update()
            self.assertEqual(constraint.problem, None)
            self.assertNotIn(constraint, model.constraints)
        else:
            self.assertRaises(
                optlang.exceptions.IndicatorConstraintsNotSupported,
                self.interface.Constraint,
                self.interface.Variable('chip') + self.interface.Variable('chap'),
                indicator_variable=self.interface.Variable('indicator', type='binary')
            )

    @abc.abstractmethod
    def test_get_primal(self):
        pass

    def test_get_dual(self):
        self.assertEqual(self.constraint.dual, None)
        self.model.optimize()
        self.assertEqual(self.model.status, 'optimal')
        self.assertAlmostEqual(self.model.objective.value, 0.8739215069684305)
        self.assertTrue(isinstance(self.model.constraints[0].dual, float))

    def test_change_constraint_name(self):
        constraint = self.interface.Constraint.clone(self.constraint)
        self.assertEqual(constraint.name, 'woodchips')
        constraint.name = 'ketchup'
        self.assertEqual(constraint.name, 'ketchup')
        self.assertEqual([constraint.name for constraint in self.model.constraints],
                         ['M_13dpg_c', 'M_2pg_c', 'M_3pg_c', 'M_6pgc_c', 'M_6pgl_c', 'M_ac_c', 'M_ac_e',
                          'M_acald_c', 'M_acald_e', 'M_accoa_c', 'M_acon_C_c', 'M_actp_c', 'M_adp_c', 'M_akg_c',
                          'M_akg_e', 'M_amp_c', 'M_atp_c', 'M_cit_c', 'M_co2_c', 'M_co2_e', 'M_coa_c', 'M_dhap_c',
                          'M_e4p_c', 'M_etoh_c', 'M_etoh_e', 'M_f6p_c', 'M_fdp_c', 'M_for_c', 'M_for_e', 'M_fru_e',
                          'M_fum_c', 'M_fum_e', 'M_g3p_c', 'M_g6p_c', 'M_glc_D_e', 'M_gln_L_c', 'M_gln_L_e',
                          'M_glu_L_c', 'M_glu_L_e', 'M_glx_c', 'M_h2o_c', 'M_h2o_e', 'M_h_c', 'M_h_e', 'M_icit_c',
                          'M_lac_D_c', 'M_lac_D_e', 'M_mal_L_c', 'M_mal_L_e', 'M_nad_c', 'M_nadh_c', 'M_nadp_c',
                          'M_nadph_c', 'M_nh4_c', 'M_nh4_e', 'M_o2_c', 'M_o2_e', 'M_oaa_c', 'M_pep_c', 'M_pi_c',
                          'M_pi_e', 'M_pyr_c', 'M_pyr_e', 'M_q8_c', 'M_q8h2_c', 'M_r5p_c', 'M_ru5p_D_c', 'M_s7p_c',
                          'M_succ_c', 'M_succ_e', 'M_succoa_c', 'M_xu5p_D_c'])
        for i, constraint in enumerate(self.model.constraints):
            constraint.name = 'c' + str(i)
        self.assertEqual([constraint.name for constraint in self.model.constraints],
                         ['c' + str(i) for i in range(0, len(self.model.constraints))])

    def test_setting_lower_bound_higher_than_upper_bound_raises(self):
        self.assertRaises(ValueError, setattr, self.model.constraints[0], 'lb', 10000000000.)
        self.assertRaises(ValueError, setattr, self.model.constraints[0], "ub", -1000000000.)

        self.assertRaises(ValueError, self.interface.Constraint, 0, lb=0, ub=-1)

    def test_setting_bounds(self):
        var = self.interface.Variable("test", lb=-10)
        c = self.interface.Constraint(var, lb=0)
        model = self.interface.Model()
        obj = self.interface.Objective(var)
        model.add(c)
        model.objective = obj

        c.ub = 5
        model.optimize()
        self.assertAlmostEqual(var.primal, 5)
        c.ub = 4
        model.optimize()
        self.assertAlmostEqual(var.primal, 4)
        c.lb = -3
        model.objective.direction = "min"
        model.optimize()
        self.assertAlmostEqual(var.primal, -3)
        c.lb = sympy.Number(-4)  # Sympy numbers should be valid bounds
        model.optimize()
        self.assertAlmostEqual(var.primal, -4)

    def test_setting_nonnumerical_bounds_raises(self):
        var = self.interface.Variable("test")
        constraint = self.interface.Constraint(var, lb=0)
        self.assertRaises(TypeError, setattr, constraint, "lb", "noodle soup")
        self.assertRaises(TypeError, setattr, self.model.constraints[0], 'lb', 'Chicken soup')
        self.assertRaises(TypeError, setattr, constraint, "ub", "noodle soup")
        self.assertRaises(TypeError, setattr, self.model.constraints[0], 'ub', 'Chicken soup')

    def test_set_constraint_bounds_to_none(self):
        model = self.interface.Model()
        var = self.interface.Variable("test")
        const = self.interface.Constraint(var, lb=-10, ub=10)
        obj = self.interface.Objective(var)
        model.add(const)
        model.objective = obj
        self.assertEqual(model.optimize(), interface.OPTIMAL)
        const.ub = None
        self.assertEqual(model.optimize(), interface.UNBOUNDED)
        const.ub = 10
        const.lb = None
        obj.direction = "min"
        self.assertEqual(model.optimize(), interface.UNBOUNDED)
        const.lb = -10
        self.assertEqual(model.optimize(), interface.OPTIMAL)

    def test_constraint_get_linear_coefficients_raises(self):
        self.assertRaises(Exception, self.constraint.get_linear_coefficients, [])

    def test_constraint_set_linear_coefficients_raises(self):
        self.assertRaises(Exception, self.constraint.set_linear_coefficients, {})

    def test_move_constant_to_rhs(self):
        x = self.interface.Variable("x")
        c1 = self.interface.Constraint(x + 3, lb=0, ub=0)
        self.assertEqual(c1.expression - x, 0)
        self.assertEqual(c1.lb, -3)
        self.assertEqual(c1.ub, -3)

        c2 = self.interface.Constraint(x - 3, lb=0, ub=0)
        self.assertEqual(c2.expression - x, 0)
        self.assertEqual(c2.lb, 3)
        self.assertEqual(c2.ub, 3)

        c3 = self.interface.Constraint(x - 3, lb=0)
        self.assertEqual(c3.expression - x, 0)
        self.assertEqual(c3.lb, 3)

        c4 = self.interface.Constraint(x - 3, ub=0)
        self.assertEqual(c4.expression - x, 0)
        self.assertEqual(c4.ub, 3)

        c5 = self.interface.Constraint(x + 3, lb=0)
        self.assertEqual(c5.expression - x, 0)
        self.assertEqual(c5.lb, -3)

        c6 = self.interface.Constraint(x + 3, ub=0)
        self.assertEqual(c6.expression - x, 0)
        self.assertEqual(c6.ub, -3)

    def test_invalid_name_raises(self):
        with self.assertRaises(Exception):
            self.interface.Constraint(1, name="")
        with self.assertRaises(Exception):
            self.interface.Constraint(1, name="This space")
        with self.assertRaises(Exception):
            self.interface.Constraint(1, name="This\ttab")

    def test_new_invalid_name_raises(self):
        const = self.interface.Constraint(1, name="MyConstraint")
        with self.assertRaises(Exception):
            const.name = ""
        with self.assertRaises(Exception):
            const.name = "This space"
        with self.assertRaises(Exception):
            const.name = "This\ttab"

    def test_construct_with_sloppy(self):
        x, y, z, w = self.model.variables[:4]
        const = self.interface.Constraint(
            symbolics.add([symbolics.mul(symbolics.One, var) for var in [x, y, z]]),
            lb=0,
            sloppy=True
        )
        self.model.add(const)
        self.model.update()

        self.assertTrue(const.get_linear_coefficients([x, y, z, w]) == {x: 1, y: 1, z: 1, w: 0})


@six.add_metaclass(abc.ABCMeta)
class AbstractObjectiveTestCase(unittest.TestCase):
    @abc.abstractmethod
    def setUp(self):
        pass

    @abc.abstractmethod
    def test_change_direction(self):
        pass

    def test_objective_get_linear_coefficients_raises(self):
        objective = self.interface.Objective(0)
        self.assertRaises(Exception, objective.get_linear_coefficients, [])

    def test_objective_set_linear_coefficients_raises(self):
        objective = self.interface.Objective(0)
        self.assertRaises(Exception, objective.set_linear_coefficients, {})

    def test_objective_value_is_none(self):
        objective = self.interface.Objective(0)
        self.assertIs(objective.value, None)

    def test_invalid_name_raises(self):
        with self.assertRaises(Exception):
            self.interface.Objective(1, name="")
        with self.assertRaises(Exception):
            self.interface.Objective(1, name="This space")
        with self.assertRaises(Exception):
            self.interface.Objective(1, name="This\ttab")

    def test_new_invalid_name_raises(self):
        obj = self.interface.Objective(1, name="MyObjective")
        with self.assertRaises(Exception):
            obj.name = ""
        with self.assertRaises(Exception):
            obj.name = "This space"
        with self.assertRaises(Exception):
            obj.name = "This\ttab"

    def test_construct_with_sloppy(self):
        x, y, z, w = self.model.variables[:4]
        obj = self.interface.Objective(
            symbolics.add([symbolics.mul((symbolics.One, var)) for var in [x, y, z]]),
            direction="min",
            sloppy=True
        )
        self.model.objective = obj

        self.assertTrue(obj.get_linear_coefficients([x, y, z, w]) == {x: 1, y: 1, z: 1, w: 0})


@six.add_metaclass(abc.ABCMeta)
class AbstractModelTestCase(unittest.TestCase):

    def setUp(self):
        with open(TESTMODELPATH) as infile:
            self.model = self.interface.Model.from_json(json.load(infile))

    def test_create_empty_model(self):
        model = self.interface.Model()
        self.assertEqual(len(model.constraints), 0)
        self.assertEqual(len(model.variables), 0)
        self.assertEqual(model.objective.expression - 0, 0)

    @abc.abstractmethod
    def test_pickle_ability(self):
        pass

    def test_pickle_empty_model(self):
        model = self.interface.Model()
        self.assertEqual(model.objective.expression - 0, 0)
        self.assertEqual(len(model.variables), 0)
        self.assertEqual(len(model.constraints), 0)
        pickle_string = pickle.dumps(model)
        from_pickle = pickle.loads(pickle_string)
        self.assertEqual(from_pickle.objective.expression - 0, 0)
        self.assertEqual(len(from_pickle.variables), 0)
        self.assertEqual(len(from_pickle.constraints), 0)

    def test_copy(self):
        self.model.optimize()
        value = self.model.objective.value
        model_copy = copy.copy(self.model)
        self.assertIsNot(self.model, model_copy)
        model_copy.optimize()
        self.assertAlmostEqual(value, model_copy.objective.value)
        self.assertEqual([(var.lb, var.ub, var.name, var.type) for var in model_copy.variables.values()],
                         [(var.lb, var.ub, var.name, var.type) for var in self.model.variables.values()])
        self.assertEqual([(constr.lb, constr.ub, constr.name) for constr in model_copy.constraints],
                         [(constr.lb, constr.ub, constr.name) for constr in self.model.constraints])

    def test_deepcopy(self):
        self.model.optimize()
        value = self.model.objective.value
        model_copy = copy.deepcopy(self.model)
        self.assertIsNot(self.model, model_copy)
        self.assertIsNot(self.model.problem, model_copy.problem)
        model_copy.optimize()
        self.assertAlmostEqual(value, model_copy.objective.value)
        self.assertEqual([(var.lb, var.ub, var.name, var.type) for var in model_copy.variables.values()],
                         [(var.lb, var.ub, var.name, var.type) for var in self.model.variables.values()])
        self.assertEqual([(constr.lb, constr.ub, constr.name) for constr in model_copy.constraints],
                         [(constr.lb, constr.ub, constr.name) for constr in self.model.constraints])

    @abc.abstractmethod
    def test_config_gets_copied_too(self):
        pass

    @abc.abstractmethod
    def test_init_from_existing_problem(self):
        pass

    def test_add_variable(self):
        model = self.interface.Model()
        var = self.interface.Variable('x')
        self.assertEqual(var.problem, None)
        model.add(var)
        self.assertTrue(var in model.variables)
        self.assertEqual(model.variables['x'].problem, var.problem)
        self.assertEqual(model.variables['x'].problem, model)
        var = self.interface.Variable('y', lb=-13)
        model.add(var)
        self.assertTrue(var in model.variables)
        self.assertEqual(model.variables['x'].lb, None)
        self.assertEqual(model.variables['x'].ub, None)
        self.assertEqual(model.variables['y'].lb, -13)
        self.assertEqual(model.variables['x'].ub, None)

    def test_add_integer_var(self):
        var = self.interface.Variable('int_var', lb=-13, ub=499., type='integer')
        self.model.add(var)
        self.assertEqual(self.model.variables['int_var'].type, 'integer')
        self.assertEqual(self.model.variables['int_var'].ub, 499.)
        self.assertEqual(self.model.variables['int_var'].lb, -13)

    @abc.abstractmethod
    def test_add_non_cplex_conform_variable(self):
        pass

    def test_remove_variable(self):
        var = self.model.variables[0]
        self.assertEqual(var.problem, self.model)
        self.model.remove(var)
        self.assertNotIn(var, self.model.variables)
        self.assertEqual(var.problem, None)

    def test_remove_variable_str(self):
        var = self.model.variables.values()[0]
        self.model.remove(var.name)
        self.assertNotIn(var, self.model.variables)
        self.assertEqual(var.problem, None)

    def test_add_constraints(self):
        x = self.interface.Variable('x', type='binary')
        y = self.interface.Variable('y', lb=-181133.3, ub=12000., type='continuous')
        z = self.interface.Variable('z', lb=0., ub=3, type='integer')
        constr1 = self.interface.Constraint(0.3 * x + 0.4 * y + 66. * z, lb=-100, ub=0., name='test')
        constr2 = self.interface.Constraint(2.333 * x + y + 3.333, ub=100.33, name='test2')
        constr3 = self.interface.Constraint(2.333 * x + y + z, ub=100.33, lb=-300)
        constr4 = self.interface.Constraint(77 * x, lb=10, name='Mul_constraint')
        constr5 = self.interface.Constraint(x, ub=-10, name='Only_var_constraint')
        constr6 = self.interface.Constraint(3, ub=88., name='Number_constraint')
        self.model.add(constr1)
        self.model.update()
        self.model.add(constr2)
        self.model.update()
        self.model.add(constr3, sloppy=True)
        self.model.update()
        self.model.add([constr4, constr5, constr6])
        self.model.update()
        self.assertIn(constr1.name, self.model.constraints)
        self.assertIn(constr2.name, self.model.constraints)
        self.assertIn(constr3.name, self.model.constraints)
        self.assertIn(constr4.name, self.model.constraints)
        self.assertIn(constr5.name, self.model.constraints)
        self.assertIn(constr6.name, self.model.constraints)

    def test_remove_constraints(self):
        x = self.interface.Variable('x', type='binary')
        y = self.interface.Variable('y', lb=-181133.3, ub=12000., type='continuous')
        z = self.interface.Variable('z', lb=4, ub=4, type='integer')
        constr1 = self.interface.Constraint(0.3 * x + 0.4 * y + 66. * z, lb=-100, ub=0., name='test')
        self.assertEqual(constr1.problem, None)
        self.model.add(constr1)
        self.model.update()
        self.assertEqual(constr1.problem, self.model)
        self.assertIn(constr1, self.model.constraints)
        self.model.remove(constr1.name)
        self.model.update()
        self.assertEqual(constr1.problem, None)
        self.assertNotIn(constr1, self.model.constraints)

    def test_add_nonlinear_constraint_raises(self):
        x = self.interface.Variable('x', type='binary')
        y = self.interface.Variable('y', lb=-181133.3, ub=12000., type='continuous')
        z = self.interface.Variable('z', lb=3, ub=3, type='integer')
        with self.assertRaises(ValueError):
            constraint = self.interface.Constraint(0.3 * x + 0.4 * y ** x + 66. * z, lb=-100, ub=0., name='test')
            self.model.add(constraint)
            self.model.update()

    def test_objective_get_linear_coefficients(self):
        coefs = self.model.objective.get_linear_coefficients(self.model.variables)
        self.assertEqual(len(coefs), len(self.model.variables))
        expr = sum(c * v for v, c in coefs.items())
        self.assertEqual((expr - self.model.objective.expression).expand() - 0, 0)

    def test_constraint_get_linear_coefficients(self):
        constraint = self.model.constraints[5]
        coefs = constraint.get_linear_coefficients(self.model.variables)
        self.assertEqual(len(coefs), len(self.model.variables))
        expr = sum(c * v for v, c in coefs.items())
        self.assertEqual((expr - constraint.expression).expand() - 0, 0)

    @abc.abstractmethod
    def test_change_of_constraint_is_reflected_in_low_level_solver(self):
        pass

    @abc.abstractmethod
    def test_constraint_set_problem_to_None_caches_the_latest_expression_from_solver_instance(self):
        pass

    @abc.abstractmethod
    def test_change_of_objective_is_reflected_in_low_level_solver(self):
        pass

    @abc.abstractmethod
    def test_change_variable_bounds(self):
        pass

    def test_change_variable_type(self):
        self.model.variables[-1].type = "integer"
        self.assertEqual(self.model.variables[-1].type, "integer")

    @abc.abstractmethod
    def test_change_constraint_bounds(self):
        pass

    def test_initial_objective(self):
        self.assertEqual(self.model.objective.expression, 1.0 * self.model.variables["R_Biomass_Ecoli_core_w_GAM"])

    def test_set_objective_in_constructor(self):
        var = self.interface.Variable("x", ub=5)
        obj = self.interface.Objective(var, direction="max")
        model = self.interface.Model(objective=obj, variables=[var])
        model.optimize()
        self.assertIs(model.objective, obj)
        self.assertAlmostEqual(model.objective.value, 5)


    def test_optimize(self):
        self.model.optimize()
        self.assertEqual(self.model.status, 'optimal')
        self.assertAlmostEqual(self.model.objective.value, 0.8739215069684303)

    def test_optimize_milp(self):
        with open(TESTMILPMODELPATH) as infile:
            milp_model = self.interface.Model.from_json(json.load(infile))
        milp_model.optimize()
        self.assertEqual(milp_model.status, 'optimal')
        self.assertAlmostEqual(milp_model.objective.value, 122.5)
        for variable in milp_model.variables:
            if variable.type == 'integer':
                self.assertEqual(variable.primal % 1, 0)

    def test_change_objective(self):
        v1, v2 = self.model.variables.values()[0:2]
        self.model.objective = self.interface.Objective(1. * v1 + 1. * v2)
        self.assertEqual(self.model.objective.direction, "max")
        self.assertEqual(
            (self.model.objective.expression -
             (1.0 * self.model.variables["R_PGK"] + 1.0 * self.model.variables["R_Biomass_Ecoli_core_w_GAM"])).expand(),
            0.
        )
        self.model.objective = self.interface.Objective(v1 + v2)
        self.assertEqual(self.model.objective.direction, "max")
        self.assertEqual(
            (self.model.objective.expression -
             (1.0 * self.model.variables["R_PGK"] + 1.0 * self.model.variables["R_Biomass_Ecoli_core_w_GAM"])).expand(),
            0.
        )

    def test_number_objective(self):
        self.model.objective = self.interface.Objective(0.)
        self.assertEqual(self.model.objective.expression - 0, 0)
        self.assertEqual(self.model.objective.direction, "max")
        self.assertEqual(self.model.optimize(), "optimal")

    def test_raise_on_non_linear_objective(self):
        """Test that an exception is raised when a non-linear objective is added to the model."""
        v1, v2 = self.model.variables.values()[0:2]
        self.assertRaises(ValueError, self.interface.Objective, v1 * v2 ** 3)

    @abc.abstractmethod
    def test_iadd_objective(self):
        pass

    @abc.abstractmethod
    def test_imul_objective(self):
        pass

    @abc.abstractmethod
    def test_set_copied_objective(self):
        pass

    @abc.abstractmethod
    def test_timeout(self):
        pass

    @abc.abstractmethod
    def test_set_linear_coefficients_objective(self):
        pass

    def test_instantiating_model_with_different_solver_problem_raises(self):
        self.assertRaises(TypeError, self.interface.Model, problem='Chicken soup')

    @abc.abstractmethod
    def test_set_linear_coefficients_constraint(self):
        pass

    def test_primal_values(self):
        self.model.optimize()
        primals = self.model.primal_values
        for var in self.model.variables:
            self.assertEqual(var.primal, primals[var.name])
        self.assertEqual(set(var.name for var in self.model.variables), set(primals))

    def test_reduced_costs(self):
        self.model.optimize()
        reduced_costs = self.model.reduced_costs
        for var in self.model.variables:
            self.assertEqual(var.dual, reduced_costs[var.name])
        self.assertEqual(set(var.name for var in self.model.variables), set(reduced_costs))

    def test_dual_values(self):
        self.model.optimize()
        constraint_primals = self.model.constraint_values  # TODO Fix this method name
        for constraint in self.model.constraints:
            self.assertEqual(constraint.primal, constraint_primals[constraint.name])
        self.assertEqual(set(const.name for const in self.model.constraints), set(constraint_primals))

    def test_shadow_prices(self):
        self.model.optimize()
        shadow_prices = self.model.shadow_prices
        for constraint in self.model.constraints:
            self.assertEqual(constraint.dual, shadow_prices[constraint.name])
        self.assertEqual(set(const.name for const in self.model.constraints), set(shadow_prices))

    def test_change_objective_can_handle_removed_vars(self):
        self.model.objective = self.interface.Objective(self.model.variables[0])
        self.model.remove(self.model.variables[0])
        self.model.update()
        self.model.objective = self.interface.Objective(self.model.variables[1] * 2)

    def test_clone_model_with_json(self):
        self.assertEqual(self.model.configuration.verbosity, 0)
        self.model.configuration.verbosity = 3
        self.model.optimize()
        opt = self.model.objective.value
        cloned_model = self.interface.Model.clone(self.model)
        self.assertEqual(cloned_model.configuration.verbosity, 3)
        self.assertEqual(len(cloned_model.variables), len(self.model.variables))
        self.assertEqual(len(cloned_model.constraints), len(self.model.constraints))
        cloned_model.optimize()
        self.assertAlmostEqual(cloned_model.objective.value, opt)

    def test_clone_model_with_lp(self):
        self.assertEqual(self.model.configuration.verbosity, 0)
        self.model.configuration.verbosity = 3
        self.model.optimize()
        opt = self.model.objective.value
        cloned_model = self.interface.Model.clone(self.model, use_lp=True)
        self.assertEqual(cloned_model.configuration.verbosity, 3)
        self.assertEqual(len(cloned_model.variables), len(self.model.variables))
        self.assertEqual(len(cloned_model.constraints), len(self.model.constraints))
        cloned_model.optimize()
        self.assertAlmostEqual(cloned_model.objective.value, opt)

    def test_clone_model_without_json(self):
        self.assertEqual(self.model.configuration.verbosity, 0)
        self.model.configuration.verbosity = 3
        self.model.optimize()
        opt = self.model.objective.value
        cloned_model = self.interface.Model.clone(self.model, use_json=False)
        self.assertEqual(cloned_model.configuration.verbosity, 3)
        self.assertEqual(len(cloned_model.variables), len(self.model.variables))
        self.assertEqual(len(cloned_model.constraints), len(self.model.constraints))
        cloned_model.optimize()
        print(cloned_model.objective.expression)
        print(cloned_model.objective.value)
        self.assertAlmostEqual(cloned_model.objective.value, opt)

    def test_remove_variable_not_in_model_raises(self):
        var = self.interface.Variable("test")
        self.assertRaises(Exception, self.model._remove_variables, [var])

    def test_objective_set_linear_coefficients(self):
        x = self.interface.Variable("x", lb=0)
        y = self.interface.Variable("y", lb=0)
        c1 = self.interface.Constraint((y + 2 * (x - 3)).expand(), ub=0)
        c2 = self.interface.Constraint(y + (1 / 2) * x - 3, ub=0)
        obj = self.interface.Objective(x)
        model = self.interface.Model()
        model.add([c1, c2])
        model.objective = obj

        self.assertEqual(model.optimize(), optlang.interface.OPTIMAL)
        self.assertAlmostEqual(x.primal, 3)
        self.assertAlmostEqual(y.primal, 0)

        obj.set_linear_coefficients({y: 1})
        self.assertEqual(float((obj.expression - (x + y)).expand()), 0.0)
        self.assertEqual(model.optimize(), optlang.interface.OPTIMAL)
        self.assertAlmostEqual(x.primal, 2)
        self.assertAlmostEqual(y.primal, 2)

        obj.set_linear_coefficients({x: 0})
        self.assertEqual(float((obj.expression - y).expand()), 0.0)
        self.assertEqual(model.optimize(), optlang.interface.OPTIMAL)
        self.assertAlmostEqual(x.primal, 0)
        self.assertAlmostEqual(y.primal, 3)

    def test_constraint_set_linear_coefficients(self):
        x = self.interface.Variable("x", lb=0, ub=1000)
        y = self.interface.Variable("y", lb=0)
        c1 = self.interface.Constraint(y, ub=1)
        obj = self.interface.Objective(x)
        model = self.interface.Model()
        model.add([c1])
        model.objective = obj

        self.assertEqual(model.optimize(), optlang.interface.OPTIMAL)
        self.assertAlmostEqual(x.primal, x.ub)

        c1.set_linear_coefficients({x: 1})
        self.assertEqual((c1.expression - (x + y)).expand() - 0, 0)
        self.assertEqual(model.optimize(), optlang.interface.OPTIMAL)
        self.assertAlmostEqual(x.primal, 1)

        c1.set_linear_coefficients({x: 2})
        self.assertEqual((c1.expression - (2 * x + y)).expand() - 0, 0)
        self.assertEqual(model.optimize(), optlang.interface.OPTIMAL)
        self.assertAlmostEqual(x.primal, 0.5)

    def test_objective_handles_constants_2(self):
        offset = 2
        self.model.optimize()
        optimal = self.model.objective.value
        objective = self.interface.Objective(self.model.objective.expression + offset)
        self.model.objective = objective
        self.model.optimize()
        self.assertAlmostEqual(self.model.objective.value, optimal + offset)

    # def test_objective_handles_constants_3(self):
    #     offset = 3
    #     self.model.optimize()
    #     optimal = self.model.objective.value
    #     self.model.objective += offset
    #     self.model.optimize()
    #     self.assertEqual(self.model.objective.value, optimal + offset)

    def test_objective_expression_includes_constant(self):
        objective = self.model.objective
        self.model.objective = self.interface.Objective(objective.expression + 3)
        self.model.update()
        self.assertEqual((self.model.objective.expression - (objective.expression + 3.)).expand(), 0.)

    def test_is_integer(self):
        model = self.model
        self.assertFalse(model.is_integer)
        self.assertFalse(optlang.interface.Model.is_integer.fget(model))

        model.variables[0].type = "integer"
        self.assertTrue(model.is_integer)
        self.assertTrue(optlang.interface.Model.is_integer.fget(model))

        model.variables[0].type = "continuous"
        model.variables[1].type = "binary"
        self.assertTrue(model.is_integer)

        model.variables[1].type = "continuous"
        self.assertFalse(model.is_integer)

    def test_binary_variables(self):
        model = self.interface.Model()
        var = self.interface.Variable("x", type="binary")
        obj = self.interface.Objective(var)
        model.objective = obj

        for lb, ub in ((0, 0), (0, 1), (1, 1)):
            var.ub = ub
            var.lb = lb

            obj.direction = "max"
            model.optimize()
            self.assertAlmostEqual(var.primal, ub)

            obj.direction = "min"
            model.optimize()
            self.assertAlmostEqual(var.primal, lb)

    def test_integer_variable_dual(self):
        model = self.interface.Model()
        x = self.interface.Variable("x", lb=0)
        y = self.interface.Variable("y", lb=0)
        c = self.interface.Constraint(x + y, ub=1)
        model.add(c)
        model.objective = self.interface.Objective(x)

        model.optimize()
        self.assertEqual(y.dual, -1)

        x.type = "integer"
        model.optimize()
        self.assertRaises(ValueError, partial(getattr, y, "dual"))

        x.type = "continuous"
        model.optimize()
        self.assertEqual(y.dual, -1)

    def test_integer_constraint_dual(self):
        model = self.interface.Model()
        x = self.interface.Variable("x")
        c = self.interface.Constraint(x, ub=1)
        model.add(c)
        model.objective = self.interface.Objective(x)

        model.optimize()
        self.assertEqual(c.dual, 1)

        x.type = "integer"
        model.optimize()
        self.assertRaises(ValueError, partial(getattr, c, "dual"))

        x.type = "continuous"
        model.optimize()
        self.assertEqual(c.dual, 1)

    def test_integer_batch_duals(self):
        model = self.interface.Model()
        x = self.interface.Variable("x")
        c = self.interface.Constraint(x, ub=1)
        model.add(c)
        model.objective = self.interface.Objective(x)
        model.optimize()

        self.assertEqual(model.reduced_costs[x.name], 0)
        self.assertEqual(model.shadow_prices[c.name], 1)

        x.type = "integer"
        model.optimize()

        with self.assertRaises(ValueError):
            model.reduced_costs
        with self.assertRaises(ValueError):
            model.shadow_prices

        x.type = "continuous"
        model.optimize()

        self.assertEqual(model.reduced_costs[x.name], 0)
        self.assertEqual(model.shadow_prices[c.name], 1)

    def test_large_objective(self):
        model = self.interface.Model()
        model.add([self.interface.Variable(str(i), lb=1) for i in range(1100)])
        model.optimize()

        obj = self.interface.Objective(
            optlang.symbolics.add([optlang.symbolics.mul((optlang.symbolics.One, v)) for v in model.variables]),
            direction="min"
        )
        model.objective = obj
        model.optimize()
        self.assertAlmostEqual(model.objective.value, len(model.variables))

    def test_implicitly_convert_milp_to_lp(self):
        model = self.interface.Model()
        var1 = self.interface.Variable("x", ub=1)
        var2 = self.interface.Variable("y", type="integer")
        var3 = self.interface.Variable("z", type="integer")
        model.add([var1, var2])
        model.optimize()
        self.assertTrue(model.is_integer)

        var2.type = "continuous"
        model.optimize()
        self.assertAlmostEqual(model.reduced_costs["x"], 0)

        var2.type = "integer"
        model.add(var3)
        model.optimize()
        self.assertTrue(model.is_integer)

        model.remove(var2)
        model.optimize()
        model.remove(var3)
        model.optimize()
        self.assertAlmostEqual(model.reduced_costs["x"], 0)

    def test_change_constraint_name(self):
        model = self.model
        const = model.constraints[0]
        old_name = const.name
        new_name = "test"
        const.name = new_name
        self.assertEqual(const.name, new_name)
        const.name = old_name
        self.assertEqual(const.name, old_name)


@six.add_metaclass(abc.ABCMeta)
class AbstractConfigurationTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def test_tolerance_parameters(self):
        model = self.interface.Model()
        params = dir(model.configuration.tolerances)
        for param in params:
            val = getattr(model.configuration.tolerances, param)
            print(val)
            setattr(model.configuration.tolerances, param, 2 * val)
            self.assertEqual(
                getattr(model.configuration.tolerances, param), 2 * val
            )


@six.add_metaclass(abc.ABCMeta)
class AbstractQuadraticProgrammingTestCase(unittest.TestCase):
    @abc.abstractmethod
    def setUp(self):
        pass

    @abc.abstractmethod
    def test_convex_obj(self):
        pass

    @abc.abstractmethod
    def test_non_convex_obj(self):
        pass

    @abc.abstractmethod
    def test_qp_convex(self):
        pass

    @abc.abstractmethod
    def test_qp_non_convex(self):
        pass



# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import copy
import os
import pickle
import random
import unittest

try:  # noqa: C901
    import osqp
except ImportError as e:

    if str(e).find('osqp') >= 0:
        class TestMissingDependency(unittest.TestCase):

            @unittest.skip('Missing dependency - ' + str(e))
            def test_fail(self):
                pass
    else:
        raise
else:

    from optlang.tests import abstract_test_cases
    from optlang.exceptions import SolverError
    from optlang.interface import OPTIMAL, INFEASIBLE, SPECIAL

    from optlang.osqp_interface import Variable, Constraint, Model, Objective
    from optlang import interface, osqp_interface

    from numpy.testing import assert_allclose
    import numpy as np
    import json
    import six

    random.seed(666)
    TESTMODELPATH = os.path.join(os.path.dirname(__file__), 'data/model.lp')
    TESTMILPMODELPATH = os.path.join(os.path.dirname(__file__), 'data/simple_milp.lp')
    LARGE_QPS = [
        os.path.join(os.path.dirname(__file__), p)
        for p in ('data/QPLIB_8785.json', 'data/QPLIB_8938.json')]
    # taken from http://qplib.zib.de/instances.html
    LARGE_QPS = dict(zip(
        LARGE_QPS,
        ((10399, 11362, 7867.4911490000004051),
         (4001, 11999, -35.7794529499999996))))


    class VariableTestCase(abstract_test_cases.AbstractVariableTestCase):
        __test__ = True

        interface = osqp_interface

        def setUp(self):
            self.var = self.interface.Variable('test')
            dummy = self.interface.Variable('dummy')
            self.model = self.interface.Model()
            self.model.add([dummy])

        def test_get_primal(self):
            self.assertEqual(self.var.primal, None)
            with open(TESTMODELPATH) as tp:
                model = Model.from_lp(tp.read())
            print(model.problem.settings)
            model.optimize()
            self.assertEqual(model.status, 'optimal')
            assert_allclose(model.objective.value, 0.8739215069684305, 1e-4, 1e-4)
            ref_sol = [0.8739215069684306, -16.023526143167608, 16.023526143167604]
            primals = [var.primal for var in model.variables[0:3]]
            assert_allclose(primals, ref_sol, 1e-4, 1e-4)

        def test_get_dual(self):
            with open(TESTMODELPATH) as tp:
                model = Model.from_lp(tp.read())
            model.optimize()
            self.assertEqual(model.status, 'optimal')
            assert_allclose(model.objective.value, 0.8739215069684305, 1e-4, 1e-4)
            self.assertTrue(isinstance(model.variables[0].dual, float))

        def test_changing_variable_names_is_reflected_in_the_solver(self):
            with open(TESTMODELPATH) as tp:
                model = Model.from_lp(tp.read())
            for i, variable in enumerate(model.variables):
                old_name = variable.name
                variable.name = "var" + str(i)
                self.assertEqual(variable.name, "var" + str(i))
                self.assertIn("var" + str(i), model.problem.variables)
                self.assertIn("var" + str(i), model._variables_to_constraints_mapping)
                self.assertNotIn(old_name, model.problem.variables)
                self.assertNotIn(old_name, model.problem.variable_lbs)
                self.assertNotIn(old_name, model.problem.variable_ubs)
                self.assertNotIn(old_name, model._variables_to_constraints_mapping)

        def test_osqp_setting_bounds(self):
            with open(TESTMODELPATH) as tp:
                model = Model.from_lp(tp.read())
            var = model.variables[0]
            var.lb = 1
            self.assertEqual(var.lb, 1)
            model.update()
            self.assertEqual(model.problem.variable_lbs[var.name], 1)
            var.ub = 2
            self.assertEqual(var.ub, 2)
            model.update()
            self.assertEqual(model.problem.variable_ubs[var.name], 2)

        def test_set_bounds_to_none(self):
            model = self.model
            var = self.var

            model.objective = self.interface.Objective(1.0 * var)
            model.update()
            print(model.problem.variables)
            self.assertEqual(model.optimize(), interface.INFEASIBLE)
            var.ub = 10
            self.assertEqual(model.optimize(), interface.OPTIMAL)
            var.ub = None
            self.assertEqual(model.optimize(), interface.INFEASIBLE)
            self.model.objective.direction = "min"
            var.lb = -10
            self.assertEqual(model.optimize(), interface.OPTIMAL)
            var.lb = None
            self.assertEqual(model.optimize(), interface.INFEASIBLE)

        def test_set_bounds_method(self):
            var = self.interface.Variable("test", lb=-10)
            c = self.interface.Constraint(var, lb=-100)
            # OSQP needs at least two variables
            model = self.interface.Model()
            obj = self.interface.Objective(1.0 * var)
            model.add([c])
            model.objective = obj

            for lb, ub in ((1, 10), (-1, 5), (11, 12)):
                obj.direction = "max"
                var.set_bounds(lb, ub)
                model.optimize()
                self.assertAlmostEqual(var.primal, ub)
                obj.direction = "min"
                model.optimize()
                self.assertAlmostEqual(var.primal, lb)

            var.set_bounds(None, 0)
            model.optimize()
            self.assertAlmostEqual(var.primal, -100)

            obj.direction = "max"
            var.set_bounds(1, None)
            self.assertEqual(model.optimize(), interface.INFEASIBLE)

            self.assertRaises(ValueError, var.set_bounds, 2, 1)

        def test_set_wrong_type_raises(self):
            self.assertRaises(ValueError, self.interface.Variable, name="test", type="mayo")
            self.assertRaises(Exception, setattr, self.var, 'type', 'ketchup')
            self.model.add(self.var)
            self.model.update()
            self.assertRaises(ValueError, setattr, self.var, "type", "mustard")
            self.var.type = "continuous"
            self.assertEqual(self.var.type, "continuous")

        def test_change_variable_type(self):
            # not supported by OSQP
            pass



    class ConstraintTestCase(abstract_test_cases.AbstractConstraintTestCase):
        interface = osqp_interface

        def test_set_linear_coefficients(self):
            self.model.add(self.constraint)
            self.constraint.set_linear_coefficients({Variable('chip'): 33., self.model.variables.R_PGK: -33})
            coefs = {
                v.name: self.model.problem.constraint_coefs[(self.constraint.name, v.name)]
                for v in self.model.variables
                if (self.constraint.name, v.name) in self.model.problem.constraint_coefs
            }
            self.assertEqual(coefs,
                             dict([('R_PGK', -33.0), ('chap', 1.0), ('chip', 33.0)]))

        def test_get_primal(self):
            self.assertEqual(self.constraint.primal, None)
            self.model.optimize()
            self.assertEqual(self.model.status, 'optimal')
            assert_allclose(self.model.objective.value, 0.8739215069684305, 1e-3, 1e-3)
            primals = [constraint.primal for constraint in self.model.constraints]
            assert_allclose(primals, np.zeros(len(primals)), 1e-3, 1e-3)  # only equality constraints

        def test_get_dual(self):
            self.assertEqual(self.constraint.primal, None)
            self.model.optimize()
            self.assertEqual(self.model.status, 'optimal')
            assert_allclose(self.model.objective.value, 0.8739215069684305, 1e-3, 1e-3)
            duals = [constraint.dual for constraint in self.model.constraints]
            self.assertTrue(all(isinstance(d, float) for d in duals))

        def test_set_constraint_bounds_to_none(self):
            model = self.interface.Model()
            var = self.interface.Variable("test")
            const = self.interface.Constraint(var, lb=-10, ub=10)
            obj = self.interface.Objective(var)
            model.add(const)
            model.objective = obj
            self.assertEqual(model.optimize(), interface.OPTIMAL)
            const.ub = None
            self.assertEqual(model.optimize(), interface.INFEASIBLE)
            const.ub = 10
            const.lb = None
            obj.direction = "min"
            self.assertEqual(model.optimize(), interface.INFEASIBLE)


    class ObjectiveTestCase(abstract_test_cases.AbstractObjectiveTestCase):
        interface = osqp_interface

        def setUp(self):
            with open(TESTMODELPATH) as tp:
                model = Model.from_lp(tp.read())
            self.obj = model.objective
            self.model = model

        def test_change_direction(self):
            self.obj.direction = "min"
            self.assertEqual(self.obj.direction, "min")
            self.assertEqual(self.model.problem.direction, 1)

            self.obj.direction = "max"
            self.assertEqual(self.obj.direction, "max")
            self.assertEqual(self.model.problem.direction, -1)


    class ModelTestCase(abstract_test_cases.AbstractModelTestCase):
        interface = osqp_interface

        def test_change_variable_type(self):
            pass

        def test_clone_model_with_lp(self):
            pass

        def test_pickle_ability(self):
            self.model.optimize()
            value = self.model.objective.value
            pickle_string = pickle.dumps(self.model)
            from_pickle = pickle.loads(pickle_string)
            from_pickle.optimize()
            self.assertAlmostEqual(value, from_pickle.objective.value)
            self.assertEqual([(var.lb, var.ub, var.name, var.type) for var in from_pickle.variables.values()],
                             [(var.lb, var.ub, var.name, var.type) for var in self.model.variables.values()])
            self.assertEqual([(constr.lb, constr.ub, constr.name) for constr in from_pickle.constraints],
                             [(constr.lb, constr.ub, constr.name) for constr in self.model.constraints])

        def test_config_gets_copied_too(self):
            self.assertEquals(self.model.configuration.verbosity, 0)
            self.model.configuration.verbosity = 1
            model_copy = copy.copy(self.model)
            self.assertEqual(model_copy.configuration.verbosity, 1)

        def test_init_from_existing_problem(self):
            inner_prob = self.model.problem
            self.assertEqual(len(self.model.variables), len(inner_prob.variables))
            self.assertEqual(len(self.model.constraints), len(inner_prob.constraints))
            self.assertEqual(set(self.model.variables.keys()), inner_prob.variables)
            self.assertEqual(set(self.model.constraints.keys()), inner_prob.constraints)

        def test_add_non_cplex_conform_variable(self):
            var = Variable('12x!!@#5_3', lb=-666, ub=666)
            self.model.add(var)
            self.assertTrue(var in self.model.variables.values())
            self.assertIn(var.name, self.model.problem.variables)
            self.assertEqual(self.model.variables['12x!!@#5_3'].lb, -666)
            self.assertEqual(self.model.variables['12x!!@#5_3'].ub, 666)
            repickled = pickle.loads(pickle.dumps(self.model))
            print(repickled.variables)
            var_from_pickle = repickled.variables['12x!!@#5_3']
            self.assertIn(var_from_pickle.name, self.model.problem.variables)

        def test_osqp_remove_variable(self):
            var = self.model.variables[0]
            self.assertEqual(var.problem, self.model)
            self.model.remove(var)
            self.model.update()
            self.assertNotIn(var.name, self.model.problem.variables)
            self.assertNotIn(var.name, self.model.problem.variable_lbs)
            self.assertNotIn(var.name, self.model.problem.variable_ubs)
            self.assertEqual(var.problem, None)

        def test_change_of_constraint_is_reflected_in_low_level_solver(self):
            x = Variable('x', lb=-83.3, ub=1324422.)
            y = Variable('y', lb=-181133.3, ub=12000.)
            constraint = Constraint(0.3 * x + 0.4 * y, lb=-100, name='test')
            self.model.add(constraint)
            self.assertEqual(
                (self.model.constraints['test'].expression - (0.4 * y + 0.3 * x)).expand() - 0,
                0
            )
            self.assertEqual(
                [self.model.problem.constraint_coefs[k]
                    for k in [('test', 'x'), ('test', 'y')]],
                [0.3, 0.4])
            z = Variable('z', lb=3, ub=4, type='integer')
            constraint += 77. * z
            self.assertEqual(
                [self.model.problem.constraint_coefs[k]
                    for k in [('test', 'x'), ('test', 'y'), ('test', 'z')]],
                [0.3, 0.4, 77.])
            self.assertEqual(
                (self.model.constraints['test'].expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand() - 0,
                0
            )

        def test_implicitly_convert_milp_to_lp(self):
            pass

        def test_integer_batch_duals(self):
            pass

        def test_integer_constraint_dual(self):
            pass

        def test_integer_variable_dual(self):
            pass

        def test_is_integer(self):
            pass

        def test_optimize(self):
            self.model.optimize()
            self.assertEqual(self.model.status, 'optimal')
            assert_allclose(self.model.objective.value, 0.8739215069684303, 1e-3, 1e-3)

        def test_optimize_milp(self):
            pass

        def test_non_convex_obj(self):
            pass

        def test_constraint_set_problem_to_None_caches_the_latest_expression_from_solver_instance(self):
            x = Variable('x', lb=-83.3, ub=1324422.)
            y = Variable('y', lb=-181133.3, ub=12000.)
            constraint = Constraint(0.3 * x + 0.4 * y, lb=-100, name='test')
            self.model.add(constraint)
            z = Variable('z', lb=2, ub=5, type='integer')
            constraint += 77. * z
            self.model.remove(constraint)
            self.assertEqual(constraint.name, "test")
            self.assertEqual(constraint.lb, -100)
            self.assertEqual(constraint.ub, None)
            self.assertEqual((constraint.expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand(), 0.0)

        def test_change_of_objective_is_reflected_in_low_level_solver(self):
            x = Variable('x', lb=-83.3, ub=1324422.)
            y = Variable('y', lb=-181133.3, ub=12000.)
            objective = Objective(0.3 * x + 0.4 * y, name='obj', direction='max')
            self.model.objective = objective
            for variable in self.model.variables:
                coeff = self.model.problem.obj_linear_coefs.get(variable.name, 0.0)
                if variable.name == 'x':
                    self.assertEqual(coeff, 0.3)
                elif variable.name == 'y':
                    self.assertEqual(coeff, 0.4)
                else:
                    self.assertEqual(coeff, 0.)
            z = Variable('z', lb=0.000003, ub=0.000003, type='continuous')
            objective += 77. * z
            for variable in self.model.variables:
                coeff = self.model.problem.obj_linear_coefs.get(variable.name, 0.0)
                if variable.name == 'x':
                    self.assertEqual(coeff, 0.3)
                elif variable.name == 'y':
                    self.assertEqual(coeff, 0.4)
                elif variable.name == 'z':
                    self.assertEqual(coeff, 77.)
                else:
                    self.assertEqual(coeff, 0.)

        def test_change_variable_bounds(self):
            inner_prob = self.model.problem
            inner_problem_bounds = [
                (inner_prob.variable_lbs[v.name],
                 inner_prob.variable_ubs[v.name])
                for v in self.model.variables
            ]
            bounds = [(var.lb, var.ub) for var in self.model.variables.values()]
            self.assertEqual(bounds, inner_problem_bounds)
            for var in self.model.variables.values():
                var.lb = random.uniform(-1000, 1000)
                var.ub = random.uniform(var.lb, 1000)
            self.model.update()
            inner_problem_bounds_new = [
                (inner_prob.variable_lbs[v.name],
                 inner_prob.variable_ubs[v.name])
                for v in self.model.variables
            ]
            bounds_new = [(var.lb, var.ub) for var in self.model.variables.values()]
            self.assertNotEqual(bounds, bounds_new)
            self.assertNotEqual(inner_problem_bounds, inner_problem_bounds_new)
            self.assertEqual(bounds_new, inner_problem_bounds_new)

        def test_change_constraint_bounds(self):
            constraint = self.model.constraints[0]
            value = 42
            constraint.ub = value
            self.assertEqual(constraint.ub, value)
            constraint.lb = value
            self.assertEqual(constraint.lb, value)
            self.assertEqual(self.model.problem.constraint_lbs[constraint.name], value)
            self.assertEqual(self.model.problem.constraint_ubs[constraint.name], value)

        def test_iadd_objective(self):
            v1, v2, v3 = list(self.model.variables.values())[0:3]
            self.model.objective += 2. * v2 - 3. * v3
            obj_coeff = self.model.problem.obj_linear_coefs
            self.assertEqual(obj_coeff,
                             {v1.name: 1, v2.name: 2, v3.name: -3})

        def test_imul_objective(self):
            self.model.objective *= 2.
            obj_coeff = self.model.problem.obj_linear_coefs
            self.assertEqual(obj_coeff, {self.model.variables[0].name: 2.0})

        def test_set_copied_objective(self):
            obj_copy = copy.copy(self.model.objective)
            self.model.objective = obj_copy
            self.assertEqual(self.model.objective.direction, "max")
            self.assertEqual(self.model.objective.expression, 1.0 * self.model.variables["R_Biomass_Ecoli_core_w_GAM"])

        def test_timeout(self):
            self.model.configuration.timeout = 1e-6
            status = self.model.optimize()
            self.assertEqual(status, 'time_limit')

        def test_set_linear_coefficients_objective(self):
            self.model.objective.set_linear_coefficients({self.model.variables.R_TPI: 666.})
            self.assertEqual(self.model.problem.obj_linear_coefs[self.model.variables.R_TPI.name], 666.)

        def test_set_linear_coefficients_constraint(self):
            constraint = self.model.constraints.M_atp_c
            coeff_dict = constraint.expression.as_coefficients_dict()
            self.assertEqual(coeff_dict[self.model.variables.R_Biomass_Ecoli_core_w_GAM], -59.8100000000000)
            constraint.set_linear_coefficients({self.model.variables.R_Biomass_Ecoli_core_w_GAM: 666.})
            coeff_dict = constraint.expression.as_coefficients_dict()
            self.assertEqual(coeff_dict[self.model.variables.R_Biomass_Ecoli_core_w_GAM], 666.)

        def test_osqp_change_objective_can_handle_removed_vars(self):
            self.model.objective = Objective(self.model.variables[0])
            self.model.remove(self.model.variables[0])
            self.model.update()
            self.model.objective = Objective(self.model.variables[1] ** 2)
            self.model.remove(self.model.variables[1])
            self.model.update()
            self.model.objective = Objective(self.model.variables[2])
            self.assertEqual(self.model.problem.obj_linear_coefs,
                {self.model.variables[2].name: 1.0})


    class ConfigurationTestCase(abstract_test_cases.AbstractConfigurationTestCase):

        interface = osqp_interface

        def setUp(self):
            self.model = Model()
            self.configuration = self.model.configuration

        def test_tolerance_parameters(self):
            model = self.interface.Model()
            params = ["optimality", "feasibility"]
            for param in params:
                val = getattr(model.configuration.tolerances, param)
                print(val)
                setattr(model.configuration.tolerances, param, 2 * val)
                self.assertEqual(
                    getattr(model.configuration.tolerances, param), 2 * val
                )

        def test_solver(self):
            for option in ("qdldl", "mkl pardiso"):
                self.configuration.linear_solver = option
                self.assertEqual(self.configuration.linear_solver, option)
                self.assertEqual(self.model.problem.settings["linsys_solver"], option)

            self.assertRaises(ValueError, setattr, self.configuration, "lp_method", "weird_stuff")

        def test_lp_method(self):
            for option in ("auto", "primal"):
                self.configuration.lp_method = option
                self.assertEqual(self.configuration.lp_method, "primal")

            self.assertRaises(ValueError, setattr, self.configuration, "lp_method", "weird_stuff")

        def test_qp_method(self):
            for option in osqp_interface._QP_METHODS:
                self.configuration.qp_method = option
                self.assertEqual(self.configuration.qp_method, "primal")

            self.assertRaises(ValueError, setattr, self.configuration, "qp_method", "weird_stuff")

        def test_verbosity(self):
            for i in range(4):
                self.model.configuration.verbosity = i
                self.assertEqual(self.model.configuration.verbosity, i)

        def test_presolve(self):
            for presolve in (True, False):
                self.configuration.presolve = presolve
                self.assertEqual(self.configuration.presolve, presolve)

            self.assertRaises(ValueError, setattr, self.configuration, "presolve", "what?")


    class QuadraticProgrammingTestCase(abstract_test_cases.AbstractQuadraticProgrammingTestCase):
        def setUp(self):
            self.model = Model()
            self.x1 = Variable("x1", lb=0)
            self.x2 = Variable("x2", lb=0)
            self.c1 = Constraint(self.x1 + self.x2, lb=1)
            self.model.add([self.x1, self.x2, self.c1])

        def test_convex_obj(self):
            model = self.model
            obj = Objective(self.x1 ** 2 + self.x2 ** 2, direction="min")
            model.objective = obj
            model.optimize()
            assert_allclose(model.objective.value, 0.5, 1e-4, 1e-4)
            assert_allclose(self.x1.primal, 0.5, 1e-4, 1e-4)
            assert_allclose(self.x2.primal, 0.5, 1e-4, 1e-4)

            obj_2 = Objective(self.x1, direction="min")
            model.objective = obj_2
            model.optimize()
            assert_allclose(model.objective.value, 0.0, 1e-4, 1e-4)
            assert_allclose(self.x1.primal, 0.0, 1e-4, 1e-4)

        def test_non_convex_obj(self):
            model = self.model
            obj = Objective(self.x1 ** 2 + self.x2 ** 2, direction="max")
            model.objective = obj
            self.assertRaises(ValueError, model.optimize)

            obj_2 = Objective(self.x1, direction="min")
            model.objective = obj_2
            model.optimize()
            assert_allclose(model.objective.value, 0.0, 1e-4, 1e-4)
            assert_allclose(self.x1.primal, 0.0, 1e-4, 1e-4)

        def test_qp_convex(self):
            for qp, info in six.iteritems(LARGE_QPS):
                nv, nc, ref_sol = info
                prob = json.load(open(qp))
                model = Model.from_json(prob)
                model.configuration.tolerances.optimality = 1e-4
                model.configuration.verbosity = 1
                model.optimize()
                self.assertEqual(len(model.variables), nv)
                self.assertEqual(len(model.constraints), nc)
                self.assertEqual(model.status, OPTIMAL)
                assert_allclose(model.objective.value, ref_sol, 1e-3, 1e-3)

        def test_qp_non_convex(self):
            # unsupported by OSQP
            pass

        def test_quadratic_objective_expression(self):
            objective = Objective(self.x1 ** 2 + self.x2 ** 2, direction="min")
            self.model.objective = objective
            self.assertEqual((self.model.objective.expression - (self.x1 ** 2 + self.x2 ** 2)).simplify(), 0)
            self.assertEqual(len(self.model.problem.obj_quadratic_coefs), 2)


    class UnsolvedTestCase(unittest.TestCase):

        interface = osqp_interface

        def setUp(self):
            model = self.interface.Model()
            x = self.interface.Variable('x', lb=0, ub=10)
            constr1 = self.interface.Constraint(1. * x, lb=3, name="constr1")
            obj = self.interface.Objective(2 * x)
            model.add(x)
            model.add(constr1)
            model.objective = obj
            self.model = model
            self.continuous_var = x
            self.constraint = constr1

        def test_status(self):
            self.assertIs(self.model.status, None)

        def test_objective_value(self):
            self.assertIs(self.model.objective.value, None)

        def test_variable_primal(self):
            self.assertIs(self.continuous_var.primal, None)

        def test_variable_dual(self):
            self.assertIs(self.continuous_var.primal, None)

        def test_constraint_primal(self):
            self.assertIs(self.constraint.primal, None)

        def test_constraint_dual(self):
            self.assertIs(self.constraint.dual, None)

        def test_primal_values(self):
            with self.assertRaises(SolverError) as context:
                self.model.primal_values
            self.assertIn("not been solved", str(context.exception))

        def test_constraint_values(self):
            with self.assertRaises(SolverError) as context:
                self.model.constraint_values
            self.assertIn("not been solved", str(context.exception))

        def test_reduced_costs(self):
            with self.assertRaises(SolverError) as context:
                self.model.reduced_costs
            self.assertIn("not been solved", str(context.exception))

        def test_shadow_prices(self):
            with self.assertRaises(SolverError) as context:
                self.model.shadow_prices
            self.assertIn("not been solved", str(context.exception))


if __name__ == '__main__':
    nose.runmodule()



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Utility functions for optlang."""

import logging

import os

log = logging.getLogger(__name__)
import tempfile
import inspect
from subprocess import check_output
from optlang import symbolics
from optlang.symbolics import mul, add, Pow


def solve_with_glpsol(glp_prob):
    """Solve glpk problem with glpsol commandline solver. Mainly for testing purposes.

    # Examples
    # --------

    # >>> problem = glp_create_prob()
    # ... glp_read_lp(problem, None, "../tests/data/model.lp")
    # ... solution = solve_with_glpsol(problem)
    # ... print 'asdf'
    # 'asdf'
    # >>> print solution
    # 0.839784

    # Returns
    # -------
    # dict
    #     A dictionary containing the objective value (key ='objval')
    #     and variable primals.
    """
    from swiglpk import glp_get_row_name, glp_get_col_name, glp_write_lp, glp_get_num_rows, glp_get_num_cols

    row_ids = [glp_get_row_name(glp_prob, i) for i in range(1, glp_get_num_rows(glp_prob) + 1)]

    col_ids = [glp_get_col_name(glp_prob, i) for i in range(1, glp_get_num_cols(glp_prob) + 1)]

    with tempfile.NamedTemporaryFile(suffix=".lp", delete=True) as tmp_file:
        tmp_file_name = tmp_file.name
        glp_write_lp(glp_prob, None, tmp_file_name)
        cmd = ['glpsol', '--lp', tmp_file_name, '-w', tmp_file_name + '.sol', '--log', '/dev/null']
        term = check_output(cmd)
        log.info(term)

    try:
        with open(tmp_file_name + '.sol') as sol_handle:
            # print sol_handle.read()
            solution = dict()
            for i, line in enumerate(sol_handle.readlines()):
                if i <= 1 or line == '\n':
                    pass
                elif i <= len(row_ids):
                    solution[row_ids[i - 2]] = line.strip().split(' ')
                elif i <= len(row_ids) + len(col_ids) + 1:
                    solution[col_ids[i - 2 - len(row_ids)]] = line.strip().split(' ')
                else:
                    print(i)
                    print(line)
                    raise Exception("Argggh!")
    finally:
        os.remove(tmp_file_name + ".sol")
    return solution


def glpk_read_cplex(path):
    """Reads cplex file and returns glpk problem.

    Returns
    -------
    glp_prob
        A glpk problems (same type as returned by glp_create_prob)
    """
    from swiglpk import glp_create_prob, glp_read_lp

    problem = glp_create_prob()
    glp_read_lp(problem, None, path)
    return problem


# noinspection PyBroadException
def list_available_solvers():
    """Determine available solver interfaces (with python bindings).

    Returns
    -------
    dict
        A dict like {'GLPK': True, 'GUROBI': False, ...}
    """
    solvers = dict(GUROBI=False, GLPK=False, MOSEK=False, CPLEX=False,
                   COINOR_CBC=False, SCIPY=False, OSQP=False)
    try:
        import gurobipy

        solvers['GUROBI'] = True
        log.debug('Gurobi python bindings found at %s' % os.path.dirname(gurobipy.__file__))
    except Exception:
        log.debug('Gurobi python bindings not available.')
    try:
        import swiglpk

        solvers['GLPK'] = True
        log.debug('GLPK python bindings found at %s' % os.path.dirname(swiglpk.__file__))
    except Exception:
        log.debug('GLPK python bindings not available.')
    try:
        import mosek

        solvers['MOSEK'] = True
        log.debug('Mosek python bindings found at %s' % os.path.dirname(mosek.__file__))
    except Exception:
        log.debug('Mosek python bindings not available.')
    try:
        import cplex

        solvers['CPLEX'] = True
        log.debug('CPLEX python bindings found at %s' % os.path.dirname(cplex.__file__))
    except Exception:
        log.debug('CPLEX python bindings not available.')

    # OSQP can be provided by the base or CUDA package
    solvers['OSQP'] = False
    try:
        import cuosqp
        solvers['OSQP'] = True
        log.debug('OSQP python bindings found at %s' % os.path.dirname(cuosqp.__file__))
    except Exception:
        pass
    try:
        import osqp
        solvers['OSQP'] = True
        log.debug('OSQP python bindings found at %s' % os.path.dirname(osqp.__file__))
    except Exception:
        pass
    if not solvers['OSQP']:
        log.debug('OSQP python bindings not available.')
    try:
        import mip

        solvers['COINOR_CBC'] = True
        log.debug('COINOR_CBC python bindings found at %s' % os.path.dirname(mip.__file__))
    except Exception:
        log.debug('COINOR_CBC python bindings not available.')
    try:
        from scipy import optimize
        optimize.linprog

        solvers["SCIPY"] = True
        log.debug("Scipy linprog function found at %s" % optimize.__file__)
    except (ImportError, AttributeError):
        log.debug("Scipy solver not available")
    return solvers


def inheritdocstring(name, bases, attrs):
    """
    Use as metaclass to inherit class and method docstrings from parent.
    Adapted from http://stackoverflow.com/questions/13937500/inherit-a-parent-class-docstring-as-doc-attribute
    Use this on classes defined in solver-specific interfaces to inherit docstrings from the high-level interface.
    """
    if '__doc__' not in attrs or not attrs["__doc__"]:
        # create a temporary 'parent' to (greatly) simplify the MRO search
        temp = type('temporaryclass', bases, {})
        for cls in inspect.getmro(temp):
            if cls.__doc__ is not None:
                attrs['__doc__'] = cls.__doc__
                break

    for attr_name, attr in attrs.items():
        if not attr.__doc__:
            for cls in inspect.getmro(temp):
                try:
                    if getattr(cls, attr_name).__doc__ is not None:
                        attr.__doc__ = getattr(cls, attr_name).__doc__
                        break
                except (AttributeError, TypeError):
                    continue

    return type(name, bases, attrs)


def method_inheritdocstring(mthd):
    """Use as decorator on a method to inherit doc from parent method of same name"""
    if not mthd.__doc__:
        pass


def is_numeric(obj):
    if isinstance(obj, (int, float)) or getattr(obj, "is_Number", False):
        return True
    else:
        try:
            float(obj)
        except ValueError:
            return False
        else:
            return True


def expr_to_json(expr):
    """
    Converts a Sympy expression to a json-compatible tree-structure.
    """
    if isinstance(expr, symbolics.Mul):
        return {"type": "Mul", "args": [expr_to_json(arg) for arg in expr.args]}
    elif isinstance(expr, symbolics.Add):
        return {"type": "Add", "args": [expr_to_json(arg) for arg in expr.args]}
    elif isinstance(expr, symbolics.Symbol):
        return {"type": "Symbol", "name": expr.name}
    elif isinstance(expr, symbolics.Pow):
        return {"type": "Pow", "args": [expr_to_json(arg) for arg in expr.args]}
    elif isinstance(expr, (float, int)):
        return {"type": "Number", "value": expr}
    elif isinstance(expr, symbolics.Real):
        return {"type": "Number", "value": float(expr)}
    elif isinstance(expr, symbolics.Integer):
        return {"type": "Number", "value": int(expr)}
    else:
        raise NotImplementedError("Type not implemented: " + str(type(expr)))


def parse_expr(expr, local_dict=None):
    """
    Parses a json-object created with 'expr_to_json' into a Sympy expression.

    If a local_dict argument is passed, symbols with be looked up by name, and a new symbol will
    be created only if the name is not in local_dict.
    """
    if local_dict is None:
        local_dict = {}
    if expr["type"] == "Add":
        return add([parse_expr(arg, local_dict) for arg in expr["args"]])
    elif expr["type"] == "Mul":
        return mul([parse_expr(arg, local_dict) for arg in expr["args"]])
    elif expr["type"] == "Pow":
        return Pow(*[parse_expr(arg, local_dict) for arg in expr["args"]])
    elif expr["type"] == "Symbol":
        try:
            return local_dict[expr["name"]]
        except KeyError:
            return symbolics.Symbol(expr["name"])
    elif expr["type"] == "Number":
        return symbolics.sympify(expr["value"])
    else:
        raise NotImplementedError(expr["type"] + " is not implemented")


class TemporaryFilename(object):
    """
    Use context manager to create a temporary file that can be opened and closed, and will be deleted in the end.

    Parameters
    ----------
    suffix : str
        The file ending. Default is 'tmp'
    content : str or None
        If str, the content will be written to the file upon creation

    Example
    ----------
    >>> with TemporaryFilename() as tmp_file_name:
    >>>     with open(tmp_file_name, "w") as tmp_file:
    >>>         tmp_file.write(stuff)
    >>>     with open(tmp_file) as tmp_file:
    >>>         stuff = tmp_file.read()
    """
    def __init__(self, suffix="tmp", content=None):
        tmp_file = tempfile.NamedTemporaryFile(suffix=suffix, delete=False, mode="w")
        if content is not None:
            tmp_file.write(content)
        self.name = tmp_file.name
        tmp_file.close()

    def __enter__(self):
        return self.name

    def __exit__(self, type, value, traceback):
        os.remove(self.name)


class SolverTolerances(object):
    def __init__(self, tolerance_functions):
        self.__dict__['_functions'] = tolerance_functions

    def __getattr__(self, item):
        try:
            return self._functions[item][0]()
        except KeyError:
            raise AttributeError(item + " is not an available tolerance parameter with this solver")

    def __setattr__(self, key, value):
        if key not in self._functions:
            raise AttributeError(key + " is not an available tolerance parameter with this solver")
        self._functions[key][1](value)

    def to_dict(self):
        return {key: getattr(self, key) for key in self._functions.keys()}

    def __dir__(self):
        return list(self._functions)


if __name__ == '__main__':
    from swiglpk import glp_create_prob, glp_read_lp, glp_get_num_rows

    problem = glp_create_prob()
    glp_read_lp(problem, None, "../tests/data/model.lp")
    print("asdf", glp_get_num_rows(problem))
    solution = solve_with_glpsol(problem)
    print(solution['R_Biomass_Ecoli_core_w_GAM'])



# Copyright 2015 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SolverError(Exception):
    """Reraise solver specific errors with this unified optlang error instead."""

    def __init__(self, message, **kwargs):
        super(SolverError, self).__init__(message, **kwargs)


class ContainerAlreadyContains(Exception):
    """
    This exception is raised when the name of an object being added to a Container is already
    taken by another object.
    """
    def __init__(self, message):
        super(ContainerAlreadyContains, self).__init__(message)


class IndicatorConstraintsNotSupported(Exception):
    """
    This exception is raised when trying to add indicator variables to a constraint using a solver
    that does not support indicator variables.
    """
    def __init__(self, message):
        super(IndicatorConstraintsNotSupported, self).__init__(message)



# Copyright 2017 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains a common interface to symbolic operations in Sympy and Symengine respectively.
All symbolic operations in the optlang codebase should use these functions.
"""

from __future__ import division

import os
import six
import uuid
import logging
import optlang

logger = logging.getLogger(__name__)

# Read environment variable
SYMENGINE_PREFERENCE = os.environ.get("OPTLANG_USE_SYMENGINE", "")

if SYMENGINE_PREFERENCE.lower() in ("false", "no", "off"):
    USE_SYMENGINE = False
else:  # pragma: no cover
    try:
        import symengine
        import symengine.sympy_compat
        from symengine.sympy_compat import Symbol as symengine_Symbol
    except ImportError as e:
        if SYMENGINE_PREFERENCE.lower() in ("true", "yes", "on"):
            logger.warn("Symengine could not be imported: " + str(e))
        USE_SYMENGINE = False
    else:
        USE_SYMENGINE = True


if USE_SYMENGINE:  # pragma: no cover # noqa: C901
    import operator
    from six.moves import reduce

    optlang._USING_SYMENGINE = True

    Integer = symengine.Integer
    Real = symengine.RealDouble
    Basic = symengine.Basic
    Number = symengine.Number
    Zero = Real(0)
    One = Real(1)
    NegativeOne = Real(-1)
    sympify = symengine.sympy_compat.sympify

    Add = symengine.Add
    Mul = symengine.Mul
    Pow = symengine.sympy_compat.Pow

    class Symbol(symengine_Symbol):
        def __new__(cls, name, *args, **kwargs):
            if not isinstance(name, six.string_types):
                raise TypeError("name should be a string, not %s" % repr(type(name)))

            return symengine_Symbol.__new__(cls, name)

        def __init__(self, name, *args, **kwargs):
            super(Symbol, self).__init__(name)
            self._name = name

        def __repr__(self):
            return self._name

        def __str__(self):
            return self._name

        def __getnewargs__(self):
            return (self._name, {})

    def add(*args):
        if len(args) == 1:
            args = args[0]
        elif len(args) == 0:
            return Zero
        return Add(*args)

    def mul(*args):
        if len(args) == 1:
            args = args[0]
        elif len(args) == 0:
            return One  # if you multiply nothing the result should be zero
        return Mul(*args)

else:  # Use sympy
    import sympy
    from sympy.core.assumptions import _assume_rules
    from sympy.core.facts import FactKB
    from sympy.core.expr import Expr

    optlang._USING_SYMENGINE = False

    Integer = sympy.Integer
    Real = sympy.RealNumber
    Basic = sympy.Basic
    Number = sympy.Number
    Zero = Real(0)
    One = Real(1)
    NegativeOne = Real(-1)
    sympify = sympy.sympify

    Add = sympy.Add
    Mul = sympy.Mul
    Pow = sympy.Pow

    class Symbol(sympy.Symbol):
        def __new__(cls, name, **kwargs):
            if not isinstance(name, six.string_types):
                raise TypeError("name should be a string, not %s" % repr(type(name)))

            obj = sympy.Symbol.__new__(cls, str(uuid.uuid1()))

            obj.name = name
            obj._assumptions = FactKB(_assume_rules)
            obj._assumptions._tell('commutative', True)
            obj._assumptions._tell('uuid', uuid.uuid1())

            return obj

        def __init__(self, *args, **kwargs):
            super(Symbol, self).__init__()

    def add(*args):
        if len(args) == 1:
            args = args[0]
        elif len(args) == 0:
            return Zero
        return sympy.Add._from_args(args)

    def mul(*args):
        if len(args) == 1:
            args = args[0]
        elif len(args) == 0:
            return One
        return sympy.Mul._from_args(args)
